/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


#ifndef __DAEDALUS_ROM_H__
#define __DAEDALUS_ROM_H__


inline u32 SwapEndian( u32 x )
{
	return ((x >> 24)&0x000000FF) |
		   ((x >> 8 )&0x0000FF00) |
		   ((x << 8 )&0x00FF0000) |
		   ((x << 24)&0xFF000000);
}

class RomID
{
	public: 
		RomID() {}
		RomID( u32 crc1, u32 crc2, u8 country_id )
		{
			CRC[0] = crc1;
			CRC[1] = crc2;
			CountryID = country_id;
		}

		bool operator==( const RomID & id ) const
		{
			return CRC[0]		== id.CRC[0] &&
				   CRC[1]		== id.CRC[1] &&
				   CountryID	== id.CountryID;
		}
		
		u32		CRC[2];
		u8		CountryID;
};

#include "PushStructPack1.h"
struct ROMHeader
{
	u8		x1, x2, x3, x4;
	u32		ClockRate;
	u32		BootAddress;
	u32		Release;
	u32		CRC1;
	u32		CRC2;
	u32		Unknown0;
	u32		Unknown1;
	s8		Name[20];
	u32		Unknown2;
	u16		Unknown3;
	u8		Unknown4;
	u8		Manufacturer;
	u16		CartID;
	s8		CountryID;
	u8		Unknown5;
};
#include "PopStructPack.h"
#include "RomSettings.h"

//*****************************************************************************
//
//*****************************************************************************
enum ECicType
{
	CIC_UNKNOWN = -1,
	CIC_6101 = 0,
	CIC_6102,
	CIC_6103,
	CIC_6104,
	CIC_6105,
	CIC_6106,

	NUM_CIC_CHIPS
};


//*****************************************************************************
//
//*****************************************************************************
struct RomInfo
{
	TCHAR		szFileName[MAX_PATH+1];
	RomID		mRomID;					// The RomID (unique to this rom)

	ROMHeader	rh;						// Copy of the ROM header, correctly byteswapped
	RomSettings settings;				// Settings for this rom
	u32			nTvType;				// OS_TV_NTSC etc
};

//*****************************************************************************
// Functions
//*****************************************************************************
void ROM_ReBoot();
void ROM_Unload();
bool ROM_LoadFile(const char * filename);
bool ROM_LoadFile(const char * filename, const RomID & rom_id, const RomSettings & settings );

bool ROM_GetRomID( const char * filename, RomID * p_id, u32 * p_rom_size, ECicType * p_boot_type );
bool ROM_GetRomName( const char * filename, TCHAR * game_name );
void ROM_ResetSettings( RomSettings * p_settings );								// Reset to default

const char *	ROM_GetCountryNameFromID( u8 country_id );
u32				ROM_GetTvTypeFromID( u8 country_id );
const char *	ROM_GetSaveTypeName( ESaveType save_type );
const char *	ROM_GetCicTypeName( ECicType cic_type );

u32						ROM_GetTexureHashFrequencyAsFrames( ETextureHashFrequency thf );
ETextureHashFrequency	ROM_GetTextureHashFrequencyFromFrames( u32 frames );
const char *			ROM_GetTextureHashFrequencyDescription( ETextureHashFrequency thf );

//*****************************************************************************
// Externs (urgh)
//*****************************************************************************
extern RomInfo g_ROM;
extern u32 g_dwNumFrames;

#endif
