/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "CodeBufferManager.h"

#include "DaedMathUtil.h"

#include "Debug\DBGConsole.h"

#include "CodeGeneratorX86.h"
#include "CodeGeneratorPSP.h"

#include <psputilsforkernel.h>

class CCodeBufferManagerPSP : public CCodeBufferManager
{
public:
	CCodeBufferManagerPSP()
		:	mpBuffer( NULL )
		,	mBufferPtr( 0 )
		,	mBufferSize( 0 )
	{
	}

	virtual bool				Initialise();
	virtual void				Reset();
	virtual void				Finalise();

	virtual CCodeGenerator *	StartNewBlock();
	virtual u32					FinaliseCurrentBlock( CCodeGenerator * p_code_gen );

private:
	
	u8	*						mpBuffer;
	u32							mBufferPtr;
	u32							mBufferSize;

private:
	CAssemblyBuffer				mAssemblyBuffer;
};

//*****************************************************************************
//
//*****************************************************************************
CCodeBufferManager *	CCodeBufferManager::Create()
{
	return new CCodeBufferManagerPSP;
}

//*****************************************************************************
//
//*****************************************************************************
bool	CCodeBufferManagerPSP::Initialise()
{

	mBufferPtr = 0;
	mBufferSize = 3 * 1024 * 1024;
	mpBuffer = new u8[ mBufferSize ];

	return true;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeBufferManagerPSP::Reset()
{
	mBufferPtr = 0;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeBufferManagerPSP::Finalise()
{
	sceKernelIcacheInvalidateRange( mpBuffer, mBufferPtr );
	if (mpBuffer != NULL)
	{
		delete [] mpBuffer;
		mpBuffer = NULL;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CCodeGenerator * CCodeBufferManagerPSP::StartNewBlock()
{
	// Round up to 16 byte boundry
	u8 *	p_rounded( (u8*)AlignPow2( (u32)(mpBuffer + mBufferPtr), 64 ) );
	mBufferPtr = p_rounded - mpBuffer;
	
	//mBufferPtr = AlignPow2( mBufferPtr, 16 );	// Cache aligned

	// This is a bit of a hack. We assume that no single entry will generate more than 
	// 32k of storage. If there appear to be problems with this assumption, this
	// value can be enlarged
	if (mBufferPtr + 32768 > mBufferSize)
	{
		DAEDALUS_ERROR( "Out of memory for dynamic recompiler" );
	}

	mAssemblyBuffer.SetBuffer( mpBuffer + mBufferPtr );
	return new CCodeGeneratorPSP( &mAssemblyBuffer );
}

extern "C" { void _DaedalusICacheInvalidate( const void * address, u32 length ); } 
//*****************************************************************************
//
//*****************************************************************************
u32 CCodeBufferManagerPSP::FinaliseCurrentBlock( CCodeGenerator * p_code_gen )
{
	p_code_gen->Finalise();

	const u8 *		p_base( mAssemblyBuffer.GetStartAddress().GetTargetU8P() );
	u32				main_block_size( mAssemblyBuffer.GetSize() );

	//	DAEDALUS_ASSERT( AlignPow2( (u32)p_base, 64 ) == (u32)p_base, "Base ptr is not aligned" );
	const u8 * p_lower( RoundPointerDown( p_base, 64 ) );
	const u8 * p_upper( RoundPointerUp( p_base + main_block_size, 64 ) );
	const u32  size( p_upper - p_lower);

	//sceKernelIcacheInvalidateRange( p_lower, size );
	//sceKernelIcacheClearAll();
	_DaedalusICacheInvalidate( p_lower, size );

	mBufferPtr += main_block_size;

	return main_block_size;
}
