/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "CodeGeneratorPSP.h"
#include "N64RegisterCachePSP.h"
#include "AssemblyUtils.h"
#include "Trace.h"

#include "DaedMathUtil.h"

#include "Utility/PrintOpCode.h"

#include "Core/R4300.h"
#include "Core/R4300_Regs.h"
#include "Core/CPU.h"
#include "Core/Memory.h"
#include "Core/Registers.h"

#include "Debug/DBGConsole.h"

#include <algorithm>

using namespace AssemblyUtils;

//#define NAIVE_REGISTER_ALLOCATION
#define STUFF_TO_DO_OPTIMISATION
//#define ANALYSE_DIRECT_MEMORY_ACCESSES

#define NOT_IMPLEMENTED( x )	DAEDALUS_ERROR( x )

#ifdef ANALYSE_DIRECT_MEMORY_ACCESSES
u32		gKnownLoads = 0;
u32		gUnknownLoads = 0;
#endif

extern "C" { const void * g_ReadAddressLookupTableForDynarec = g_ReadAddressLookupTable; }

extern "C" { void _ReturnFromDynaRec(); } 
extern "C" { void _DirectExitCheck( u32 instructions_executed, u32 target_pc, u32 delay ); } 
extern "C" { void _IndirectExitCheck( u32 instructions_executed, CIndirectExitMap * p_map, u32 target_pc ); } 

#ifdef STUFF_TO_DO_OPTIMISATION
extern "C"
{
	u32 _ReadBitsDirect_u8( u32 address );
	u32 _ReadBitsDirect_s8( u32 address );
	u32 _ReadBitsDirect_u16( u32 address );
	u32 _ReadBitsDirect_s16( u32 address );
	u32 _ReadBitsDirect_u32( u32 address );

	// Dynarec calls this for simplicity
	void Write32BitsForDynaRec( u32 address, u32 value )
	{
		Write32Bits( address, value );
	}
	void Write16BitsForDynaRec( u32 address, u16 value )
	{
		Write16Bits( address, value );
	}
	void Write8BitsForDynaRec( u32 address, u8 value )
	{
		Write8Bits( address, value );
	}

	void _WriteBitsDirect_u32( u32 address, u32 value );
	void _WriteBitsDirect_u16( u32 address, u16 value );
	void _WriteBitsDirect_u8( u32 address, u8 value );
}

#define ReadBitsDirect_u8 _ReadBitsDirect_u8
#define ReadBitsDirect_s8 _ReadBitsDirect_s8
#define ReadBitsDirect_u16 _ReadBitsDirect_u16
#define ReadBitsDirect_s16 _ReadBitsDirect_s16
#define ReadBitsDirect_u32 _ReadBitsDirect_u32

#define WriteBitsDirect_u32 _WriteBitsDirect_u32
#define WriteBitsDirect_u16 _WriteBitsDirect_u16
#define WriteBitsDirect_u8 _WriteBitsDirect_u8

#else
	//*****************************************************************************
	// Read T from memory, extend to 32 bits, check for exceptions and return
	//*****************************************************************************
	template< typename T >
	u32 ReadBitsDirect( u32 address )
	{
		return u32( *(T *)FuncTableReadAddress( address ) );
	}

#define ReadBitsDirect_u8 ReadBitsDirect<u8>
#define ReadBitsDirect_s8 ReadBitsDirect<s8>
#define ReadBitsDirect_u16 ReadBitsDirect<u16>
#define ReadBitsDirect_s16 ReadBitsDirect<s16>
#define ReadBitsDirect_u32 ReadBitsDirect<u32>

#define WriteBitsDirect_u32	Write32Bits
#define WriteBitsDirect_u16	Write16Bits
#define WriteBitsDirect_u8	Write8Bits

#endif



#ifdef STUFF_TO_DO_OPTIMISATION
extern "C" { void _ReturnFromDynaRecIfStuffToDo( u32 register_mask ); } 
extern "C" { void _DaedalusICacheInvalidate( const void * address, u32 length ); } 

bool			gHaveSavedPatchedOps = false;
PspOpCode		gOriginalOps[2];
PspOpCode		gReplacementOps[2];

void * MakeUncachablePointer( void * p_in )
{
	return reinterpret_cast< u8 * >( reinterpret_cast<u32>( p_in ) | 0x40000000 );
}

void Dynarec_ClearedCPUStuffToDo()
{
	// Replace first two ops of _ReturnFromDynaRecIfStuffToDo with 'jr ra, nop'
	u8 *			p_void_function( reinterpret_cast< u8 * >( _ReturnFromDynaRecIfStuffToDo ) );
	PspOpCode *		p_function_address = reinterpret_cast< PspOpCode * >( MakeUncachablePointer( p_void_function ) );

	if(!gHaveSavedPatchedOps)
	{
		gOriginalOps[0] = p_function_address[0];
		gOriginalOps[1] = p_function_address[1];

		PspOpCode	op_code;
		op_code._u32 = 0;
		op_code.op = OP_SPECOP;
		op_code.spec_op = SpecOp_JR;
		op_code.rs = PspReg_RA;
		gReplacementOps[0] = op_code;		// JR RA
		gReplacementOps[1]._u32 = 0;		// NOP

		gHaveSavedPatchedOps = true;
	}

	p_function_address[0] = gReplacementOps[0];
	p_function_address[1] = gReplacementOps[1];

	const u8 * p_lower( RoundPointerDown( p_void_function, 64 ) );
	const u8 * p_upper( RoundPointerUp( p_void_function + 8, 64 ) );
	const u32  size( p_upper - p_lower);
	//sceKernelDcacheWritebackRange( p_lower, size );
	//sceKernelIcacheInvalidateRange( p_lower, size );

	_DaedalusICacheInvalidate( p_lower, size );
}

void Dynarec_SetCPUStuffToDo()
{
	// Restore first two ops of _ReturnFromDynaRecIfStuffToDo

	u8 *			p_void_function( reinterpret_cast< u8 * >( _ReturnFromDynaRecIfStuffToDo ) );
	PspOpCode *		p_function_address = reinterpret_cast< PspOpCode * >( MakeUncachablePointer( p_void_function ) );

	p_function_address[0] = gOriginalOps[0];
	p_function_address[1] = gOriginalOps[1];

	const u8 * p_lower( RoundPointerDown( p_void_function, 64 ) );
	const u8 * p_upper( RoundPointerUp( p_void_function + 8, 64 ) );
	const u32  size( p_upper - p_lower);
	//sceKernelDcacheWritebackRange( p_lower, size );
	//sceKernelIcacheInvalidateRange( p_lower, size );

	_DaedalusICacheInvalidate( p_lower, size );
}
#else

void Dynarec_ClearedCPUStuffToDo()
{
}
void Dynarec_SetCPUStuffToDo()
{
}

#endif

extern "C"
{
void HandleException_extern()
{
	switch (gCPUState.Delay)
	{
		case DO_DELAY:
			gCPUState.CurrentPC += 4;
			gCPUState.Delay = EXEC_DELAY;
			break;
		case EXEC_DELAY:
			gCPUState.CurrentPC = gCPUState.TargetPC;
			gCPUState.Delay = NO_DELAY;
			break;
		case NO_DELAY:
			gCPUState.CurrentPC += 4;
			break;
		default:	// MSVC extension - the default will never be reached
			NODEFAULT;
	}	
}

}

namespace
{
const EPspReg	gRegistersToUseForCaching[] = 
{
	PspReg_S0,
	PspReg_S1,
	PspReg_S2,
	PspReg_S3,
	PspReg_S4,
	PspReg_S5,
	PspReg_S6,
//	PspReg_S7,		// Used for read address table
	PspReg_T2,
	PspReg_T3,
	PspReg_T4,
	PspReg_T5,
	PspReg_T6,
	PspReg_T7,
	PspReg_T8,
	PspReg_T9,
	PspReg_A2,
	PspReg_A3,
};
}

u32		gTotalRegistersCached = 0;
u32		gTotalRegistersUncached = 0;


//*****************************************************************************
//
//*****************************************************************************
CCodeGeneratorPSP::CCodeGeneratorPSP( CAssemblyBuffer * p_buffer )
:	CCodeGenerator( )
,	CAssemblyWriterPSP( p_buffer )
,	mpBasePointer( NULL )
,	mBaseRegister( PspReg_S8 )		// TODO
{
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::Finalise()
{
	SetAssemblyBuffer( NULL );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetCachedBasePointer( const void * p_base )
{
	mpBasePointer = reinterpret_cast< const u8 * >( p_base );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetRegisterSpanList( const RegisterSpanList & span_list )
{
	mRegisterSpanList = span_list;

	// Sort in order of increasing start point
	std::sort( mRegisterSpanList.begin(), mRegisterSpanList.end(), SAscendingSpanStartSort() );

	const u32 NUM_CACHE_REGS( sizeof(gRegistersToUseForCaching) / sizeof(gRegistersToUseForCaching[0]) );

	// Push all the available registers in reverse order (i.e. use temporaries later)
	DAEDALUS_ASSERT( mAvailableRegisters.empty(), "Why isn't the available register list empty?" );
	for( s32 i = NUM_CACHE_REGS-1; i >= 0 ; --i )
	{
		mAvailableRegisters.push( gRegistersToUseForCaching[ i ] );
	}

#ifdef NAIVE_REGISTER_ALLOCATION
	u32		cache_reg_idx( 0 );
	for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		if( cache_reg_idx < NUM_CACHE_REGS )
		{
			EPspReg		cachable_reg( gRegistersToUseForCaching[cache_reg_idx] );

			mRegisterCache.SetCachedReg( span.Register, 0, cachable_reg );
			cache_reg_idx++;
		}

	}

	// If there are still unused registers, assign to high part of reg
	if( cache_reg_idx < NUM_CACHE_REGS )
	{
		for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
		{
			const SRegisterSpan &	span( *span_it );

			if( cache_reg_idx < NUM_CACHE_REGS )
			{
				EPspReg		cachable_reg( gRegistersToUseForCaching[cache_reg_idx] );

				mRegisterCache.SetCachedReg( span.Register, 1, cachable_reg );
				cache_reg_idx++;
			}
		}
	}
#endif
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::ExpireOldIntervals( u32 instruction_idx )
{
	// mActiveIntervals is held in order of increasing end point
	for(RegisterSpanList::iterator span_it = mActiveIntervals.begin(); span_it < mActiveIntervals.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		if( span.SpanEnd >= instruction_idx )
		{
			break;
		}

		// This interval is no longer active - flush the register and return it to the list of available regs
		EPspReg		psp_reg( mRegisterCache.GetCachedReg( span.Register, 0 ) );

		FlushRegister( span.Register, 0, true );

		mRegisterCache.ClearCachedReg( span.Register, 0 );

		mAvailableRegisters.push( psp_reg );

		span_it = mActiveIntervals.erase( span_it );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SpillAtInterval( const SRegisterSpan & live_span )
{
	DAEDALUS_ASSERT( !mActiveIntervals.empty(), "There are no active intervals" );

	const SRegisterSpan &	last_span( mActiveIntervals.back() );		// Spill the last active interval (it has the greatest end point)

	if( last_span.SpanEnd > live_span.SpanEnd )
	{
		// Uncache the old span
		EPspReg		psp_reg( mRegisterCache.GetCachedReg( last_span.Register, 0 ) );
		FlushRegister( last_span.Register, 0, true );
		mRegisterCache.ClearCachedReg( last_span.Register, 0 );

		// Cache the new span
		mRegisterCache.SetCachedReg( live_span.Register, 0, psp_reg );

		mActiveIntervals.pop_back();				// Remove the last span
		mActiveIntervals.push_back( live_span );	// Insert in order of increasing end point

		std::sort( mActiveIntervals.begin(), mActiveIntervals.end(), SAscendingSpanEndSort() );		// XXXX - will be quicker to insert in the correct place rather than sorting each time
	}
	else
	{
		// There is no space for this register - we just don't update the register cache info, so we save/restore it from memory as needed
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::UpdateRegisterCaching( u32 instruction_idx )
{
#ifndef NAIVE_REGISTER_ALLOCATION
	ExpireOldIntervals( instruction_idx );

	for(RegisterSpanList::const_iterator span_it = mRegisterSpanList.begin(); span_it < mRegisterSpanList.end(); ++span_it )
	{
		const SRegisterSpan &	span( *span_it );

		// As we keep the intervals sorted in order of SpanStart, we can exit as soon as we encounter a SpanStart in the future
		if( instruction_idx < span.SpanStart )
		{
			break;
		}

		// Only process live intervals
		if( instruction_idx >= span.SpanStart && instruction_idx <= span.SpanEnd )
		{
			if( !mRegisterCache.IsCached( span.Register, 0 ) )
			{
				if( mAvailableRegisters.empty() )
				{
					SpillAtInterval( span );
				}
				else
				{
					// Use this register for caching
					mRegisterCache.SetCachedReg( span.Register, 0, mAvailableRegisters.top() );

					// Pop this register from the available list
					mAvailableRegisters.pop();
					mActiveIntervals.push_back( span );		// Insert in order of increasing end point

					std::sort( mActiveIntervals.begin(), mActiveIntervals.end(), SAscendingSpanEndSort() );		// XXXX - will be quicker to insert in the correct place rather than sorting each time
				}
			}
		}
	}
#endif
}

//*****************************************************************************
//
//*****************************************************************************
RegisterSnapshotHandle	CCodeGeneratorPSP::GetRegisterSnapshot()
{
	RegisterSnapshotHandle	handle( mRegisterSnapshots.size() );
	
	mRegisterSnapshots.push_back( mRegisterCache );

	return handle;
}

//*****************************************************************************
//
//*****************************************************************************
CCodeLabel	CCodeGeneratorPSP::GetEntryPoint() const
{
	return GetAssemblyBuffer()->GetStartAddress();
}

//*****************************************************************************
//
//*****************************************************************************
u32	CCodeGeneratorPSP::GetCompiledCodeSize() const
{
	return GetAssemblyBuffer()->GetSize();
}

//*****************************************************************************
//
//*****************************************************************************
EPspReg	CCodeGeneratorPSP::GetRegisterNoLoad( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg scratch_reg )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		return mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx );
	}
	else
	{
		return scratch_reg;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetRegisterValue( EPspReg psp_reg, EN64Reg n64_reg, u32 lo_hi_idx )
{
	if( mRegisterCache.IsKnownValue( n64_reg, lo_hi_idx ) )
	{
		//printf( "Loading %s[%d] <- %08x\n", RegNames[ n64_reg ], lo_hi_idx, mRegisterCache.GetKnownValue( n64_reg, lo_hi_idx ) );
		LoadConstant( psp_reg, mRegisterCache.GetKnownValue( n64_reg, lo_hi_idx ) );
	}
	else
	{
		GetVar( psp_reg, &gGPR[ n64_reg ]._u32[ lo_hi_idx ] );
	}
}

//*****************************************************************************
//
//*****************************************************************************
EPspReg	CCodeGeneratorPSP::GetRegisterAndLoad( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg scratch_reg )
{
	EPspReg		reg;
	bool		need_load( false );

	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		gTotalRegistersCached++;
		reg = mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx );

		// We're loading it below, so set the valid flag
		if( !mRegisterCache.IsValid( n64_reg, lo_hi_idx ) )
		{
			need_load = true;
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
		}
	}
	else if( n64_reg == N64Reg_R0 )
	{
		reg = PspReg_R0;
	}
	else
	{
		gTotalRegistersUncached++;
		reg = scratch_reg;
		need_load = true;
	}

	if( need_load )
	{
		GetRegisterValue( reg, n64_reg, lo_hi_idx );
	}

	return reg;
}

//*****************************************************************************
//	Similar to GetRegisterAndLoad, but ALWAYS loads into the specified psp register
//*****************************************************************************
void CCodeGeneratorPSP::LoadRegister( EPspReg psp_reg, EN64Reg n64_reg, u32 lo_hi_idx )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		gTotalRegistersCached++;

		// Load the register if it's currently invalid
		if( !mRegisterCache.IsValid( n64_reg,lo_hi_idx ) )
		{
			GetRegisterValue( cached_reg, n64_reg, lo_hi_idx );
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
		}

		// Copy the register if necessary
		if( psp_reg != cached_reg )
		{
			OR( psp_reg, cached_reg, PspReg_R0 );
		}
	}
	else if( n64_reg == N64Reg_R0 )
	{
		OR( psp_reg, PspReg_R0, PspReg_R0 );
	}
	else
	{
		gTotalRegistersUncached++;
		GetRegisterValue( psp_reg, n64_reg, lo_hi_idx );
	}
}

//*****************************************************************************
//	This function pulls in a cached register so that it can be used at a later point.
//	This is usally done when we have a branching instruction - it guarantees that
//	the register is valid regardless of whether or not the branch is taken.
//*****************************************************************************
void	CCodeGeneratorPSP::PrepareCachedRegister( EN64Reg n64_reg, u32 lo_hi_idx )
{
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		// Load the register if it's currently invalid
		if( !mRegisterCache.IsValid( n64_reg,lo_hi_idx ) )
		{
			GetRegisterValue( cached_reg, n64_reg, lo_hi_idx );
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::StoreRegister( EN64Reg n64_reg, u32 lo_hi_idx, EPspReg psp_reg )
{
	mRegisterCache.ClearKnownValue( n64_reg, lo_hi_idx );

	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

		gTotalRegistersCached++;

		// Update our copy as necessary
		if( psp_reg != cached_reg )
		{
			OR( cached_reg, psp_reg, PspReg_R0 );
		}
		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, true );
		mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
	}
	else
	{
		gTotalRegistersUncached++;

		SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], psp_reg );

		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::SetRegister( EN64Reg n64_reg, u32 lo_hi_idx, u32 value, EPspReg scratch_reg )
{
	mRegisterCache.SetKnownValue( n64_reg, lo_hi_idx, value );
	mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, true );
	if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	{
		mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, false );		// The actual cache is invalid though!
	}

	//if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
	//{
	//	EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

	//	gTotalRegistersCached++;

	//	LoadConstant( cached_reg, value );
	//	mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, true );
	//	mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, true );
	//}
	//else
	//{
	//	EPspReg		reg;
	//	if( value == 0 )
	//	{
	//		reg = PspReg_R0;
	//	}
	//	else
	//	{
	//		LoadConstant( scratch_reg, value );
	//		reg = scratch_reg;
	//	}
	//	gTotalRegistersUncached++;

	//	SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], reg );
	//}
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::UpdateRegister( EN64Reg n64_reg, EPspReg psp_reg, EUpdateRegOptions options, EPspReg scratch_reg )
{
	StoreRegisterLo( n64_reg, psp_reg );

	switch( options )
	{
	case URO_HI_LEAVE:
		break;
	case URO_HI_SIGN_EXTEND:
		SRA( scratch_reg, psp_reg, 0x1f );		// Sign extend
		StoreRegisterHi( n64_reg, scratch_reg );
		break;
	case URO_HI_CLEAR:
		SetRegisterHi( n64_reg, 0, scratch_reg );
		break;
	case URO_HI_SET:
		SetRegisterHi( n64_reg, 0xffffffff, scratch_reg );
		break;
	default:
		DAEDALUS_ERROR( "Unhandled option" );
		NODEFAULT;
		break;
	}
}

//*****************************************************************************
//
//*****************************************************************************
EPspFloatReg	CCodeGeneratorPSP::GetFloatRegisterNoLoad( EN64FloatReg n64_reg )
{
	return EPspFloatReg( n64_reg );	// 1:1 mapping
}

//*****************************************************************************
//
//*****************************************************************************
EPspFloatReg	CCodeGeneratorPSP::GetFloatRegisterAndLoad( EN64FloatReg n64_reg )
{
	EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );	// 1:1 mapping
	if( !mRegisterCache.IsFPValid( n64_reg ) )
	{
		GetFloatVar( psp_reg, &g_qwCPR[1][n64_reg]._f32[ 0 ] );
		mRegisterCache.MarkFPAsValid( n64_reg, true );
	}

	return psp_reg;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::UpdateFloatRegister( EN64FloatReg n64_reg )
{
	mRegisterCache.MarkFPAsDirty( n64_reg, true );
	mRegisterCache.MarkFPAsValid( n64_reg, true );
}

//*****************************************************************************
//
//*****************************************************************************
const CN64RegisterCachePSP & CCodeGeneratorPSP::GetRegisterCacheFromHandle( RegisterSnapshotHandle snapshot ) const
{
	DAEDALUS_ASSERT( snapshot.Handle < mRegisterSnapshots.size(), "Invalid snapshot handle" );
	return mRegisterSnapshots[ snapshot.Handle ];
}

//*****************************************************************************
//	Flush a specific register back to memory if dirty.
//	Clears the dirty flag and invalidates the contents if specified
//*****************************************************************************
void CCodeGeneratorPSP::FlushRegister( EN64Reg n64_reg, u32 lo_hi_idx, bool invalidate )
{
	if( mRegisterCache.IsDirty( n64_reg, lo_hi_idx ) )
	{
		if( mRegisterCache.IsKnownValue( n64_reg, lo_hi_idx ) )
		{
			s32		known_value( mRegisterCache.GetKnownValue( n64_reg, lo_hi_idx ) );

			SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], known_value );
		}
		else if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsValid( n64_reg, lo_hi_idx ), "Register is dirty but not valid?" );

			EPspReg	cached_reg( mRegisterCache.GetCachedReg( n64_reg, lo_hi_idx ) );

			SetVar( &gGPR[ n64_reg ]._u32[ lo_hi_idx ], cached_reg );
		}
		else
		{
			DAEDALUS_ERROR( "Register is dirty, but not known or cached" );
		}

		// We're no longer dirty
		mRegisterCache.MarkAsDirty( n64_reg, lo_hi_idx, false );
	}

	// Invalidate the register, so we pick up any values the function might have changed
	if( invalidate )
	{
		mRegisterCache.ClearKnownValue( n64_reg, lo_hi_idx );
		if( mRegisterCache.IsCached( n64_reg, lo_hi_idx ) )
		{
			mRegisterCache.MarkAsValid( n64_reg, lo_hi_idx, false );
		}
	}
}

//*****************************************************************************
//	This function flushes all dirty registers back to memory
//	If the invalidate flag is set this also invalidates the known value/cached
//	register. This is primarily to ensure that we keep the register set
//	in a consistent set across calls to generic functions. Ideally we need
//	to reimplement generic functions with specialised code to avoid the flush.
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAllRegisters( bool invalidate )
{
	for( u32 lo_hi_idx = 0; lo_hi_idx < 2; ++lo_hi_idx )
	{
		// Skip r0
		for( u32 i = 1; i < NUM_N64_REGS; ++i )
		{
			EN64Reg	n64_reg = EN64Reg( i );

			FlushRegister( n64_reg, lo_hi_idx, invalidate );
		}
	}

	for( u32 i = 0; i < NUM_N64_FP_REGS; ++i )
	{
		EN64FloatReg	n64_reg = EN64FloatReg( i );
		if( mRegisterCache.IsFPDirty( n64_reg ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsFPValid( n64_reg ), "Register is dirty but not valid?" );

			EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );

			SetFloatVar( &g_qwCPR[1][n64_reg]._f32[ 0 ], psp_reg );

			mRegisterCache.MarkFPAsDirty( n64_reg, false );
		}

		if( invalidate )
		{
			// Invalidate the register, so we pick up any values the function might have changed
			mRegisterCache.MarkFPAsValid( n64_reg, false );
		}
	}
}

//*****************************************************************************
//	This function flushes all dirty *temporary* registers back to memory, and
//	marks them as invalid.
//*****************************************************************************
void	CCodeGeneratorPSP::FlushAllTemporaryRegisters( bool invalidate )
{
	for( u32 lo_hi_idx = 0; lo_hi_idx < 2; ++lo_hi_idx )
	{
		// Skip r0
		for( u32 i = 1; i < NUM_N64_REGS; ++i )
		{
			EN64Reg	n64_reg = EN64Reg( i );

			if( mRegisterCache.IsTemporary( n64_reg, lo_hi_idx ) )
			{
				FlushRegister( n64_reg, lo_hi_idx, invalidate );
			}
		}
	}

	// XXXX some fp regs are preserved across function calls?
	for( u32 i = 0; i < NUM_N64_FP_REGS; ++i )
	{
		EN64FloatReg	n64_reg = EN64FloatReg( i );
		if( mRegisterCache.IsFPDirty( n64_reg ) )
		{
			DAEDALUS_ASSERT( mRegisterCache.IsFPValid( n64_reg ), "Register is dirty but not valid?" );

			EPspFloatReg	psp_reg = EPspFloatReg( n64_reg );

			SetFloatVar( &g_qwCPR[1][n64_reg]._f32[ 0 ], psp_reg );

			mRegisterCache.MarkFPAsDirty( n64_reg, false );
		}

		if( invalidate )
		{
			// Invalidate the register, so we pick up any values the function might have changed
			mRegisterCache.MarkFPAsValid( n64_reg, false );
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::CallUpdateCounter( u32 num_instructions )
{
	PspOpCode	op1;
	PspOpCode	op2;

	GetLoadConstantOps( PspReg_A0, num_instructions, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( CPU_UpdateCounter ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		JAL( CCodeLabel( reinterpret_cast< const void * >( CPU_UpdateCounter ) ), false );		// No branch delay
		AppendOp( op2 );
	}
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation CCodeGeneratorPSP::GenerateExitCode( u32 exit_address, u32 jump_address, u32 num_instructions, CCodeLabel next_fragment )
{
	//DAEDALUS_ASSERT( exit_address != u32( ~0 ), "Invalid exit address" );
	DAEDALUS_ASSERT( !next_fragment.IsSet() || jump_address == 0, "Shouldn't be specifying a jump address if we have a next fragment?" );

	FlushAllRegisters( false );

	u32		exit_delay;
	if( jump_address != 0 )
	{
		SetVar( &gCPUState.TargetPC, jump_address );
		exit_delay = EXEC_DELAY;
	}
	else
	{
		exit_delay = NO_DELAY;
	}

	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, exit_address );

	CJumpLocation	jump;

	// Use the branch delay slot to load the second argument
	PspOpCode	op1;
	PspOpCode	op2;

	GetLoadConstantOps( PspReg_A2, exit_delay, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		jump = JAL( CCodeLabel( reinterpret_cast< const void * >( _DirectExitCheck ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		jump = JAL( CCodeLabel( reinterpret_cast< const void * >( _DirectExitCheck ) ), false );		// No branch delay
		AppendOp( op2 );
	}

	
	// This jump may be NULL, in which case we patch it below
	// This gets patched with a jump to the next fragment if the target is later found
	CJumpLocation jump_to_next_fragment( J( next_fragment, true ) );

	// Patch up the exit jump if the target hasn't been compiled yet
	if( !next_fragment.IsSet() )
	{
		PatchJumpLong( jump_to_next_fragment, CCodeLabel( reinterpret_cast< const void * >( _ReturnFromDynaRec ) ) );
	}
	return jump_to_next_fragment;
}

//*****************************************************************************
// Handle branching back to the interpreter after an ERET
//*****************************************************************************
void CCodeGeneratorPSP::GenerateEretExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{
	FlushAllRegisters( false );

	// Eret is a bit bodged so we exit at PC + 4
	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, reinterpret_cast< s32 >( p_map ) );
	GetVar( PspReg_A2, &gCPUState.CurrentPC );
	
	J( CCodeLabel( reinterpret_cast< const void * >( _IndirectExitCheck ) ), false );
	ADDIU( PspReg_A2, PspReg_A2, 4 );		// Delay slot
}

//*****************************************************************************
// Handle branching back to the interpreter after an indirect jump
//*****************************************************************************
void CCodeGeneratorPSP::GenerateIndirectExitCode( u32 num_instructions, CIndirectExitMap * p_map )
{
	FlushAllRegisters( false );

	// New return address is in gCPUState.TargetPC
	LoadConstant( PspReg_A0, num_instructions );
	LoadConstant( PspReg_A1, reinterpret_cast< s32 >( p_map ) );
	GetVar( PspReg_A2, &gCPUState.TargetPC );
	
	// TODO: Use branch delay slot
	J( CCodeLabel( reinterpret_cast< const void * >( _IndirectExitCheck ) ), true );
}

//*****************************************************************************
//
//*****************************************************************************
void CCodeGeneratorPSP::GenerateExceptionHander( ExceptionHandlerFn p_exception_handler_fn, std::vector< CJumpLocation > & exception_handler_jumps )
{
	CCodeLabel exception_handler( GetAssemblyBuffer()->GetLabel() );

	// XXXX - could just have J( p_fn ) here?
	JAL( CCodeLabel( reinterpret_cast< const void * >( p_exception_handler_fn ) ), true );

	J( CCodeLabel( reinterpret_cast< const void * >( _ReturnFromDynaRec ) ), true );

	for( std::vector< CJumpLocation >::iterator it = exception_handler_jumps.begin(); it != exception_handler_jumps.end(); ++it )
	{
		CJumpLocation jump( *it );
		PatchJumpLong( jump, exception_handler );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBranchHandler( CJumpLocation branch_handler_jump, RegisterSnapshotHandle snapshot )
{
	if(branch_handler_jump.IsSet())
	{
		CCodeLabel	current_label( GetAssemblyBuffer()->GetLabel() );

		PatchJumpLong( branch_handler_jump, current_label );

		mRegisterCache = GetRegisterCacheFromHandle( snapshot );
	}
	else
	{
		DAEDALUS_ERROR( "Why is the branch handler jump not set?" );
	}
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchAlways( CCodeLabel target )
{
	return J( target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfSet( const u32 * p_var, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	return BNE( PspReg_T0, PspReg_R0, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfNotSet( const u32 * p_var, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	return BEQ( PspReg_T0, PspReg_R0, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfEqual( const u32 * p_var, u32 value, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	LoadConstant( PspReg_T1, value );
	return BEQ( PspReg_T0, PspReg_T1, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfNotEqual( const u32 * p_var, u32 value, CCodeLabel target )
{
	GetVar( PspReg_T0, p_var );
	LoadConstant( PspReg_T1, value );
	return BNE( PspReg_T0, PspReg_T1, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateBranchIfNotEqual( EPspReg reg_a, u32 value, CCodeLabel target )
{
	EPspReg	reg_b( reg_a == PspReg_T0 ? PspReg_T1 : PspReg_T0 );		// Make sure we don't use the same reg

	LoadConstant( reg_b, value );
	return BNE( reg_a, reg_b, target, true );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetVar( u32 * p_var, EPspReg reg_src )
{
	DAEDALUS_ASSERT( reg_src != PspReg_AT, "Whoops, splattering source register" );

	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( p_var, &reg_base, &base_offset );

	SW( reg_src, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetVar( u32 * p_var, u32 value )
{
	EPspReg		reg_value;

	if(value == 0)
	{
		reg_value = PspReg_R0;
	}
	else
	{
		LoadConstant( PspReg_T0, value );
		reg_value = PspReg_T0;
	}

	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( p_var, &reg_base, &base_offset );

	SW( reg_value, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::SetFloatVar( f32 * p_var, EPspFloatReg reg_src )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( p_var, &reg_base, &base_offset );

	SWC1( reg_src, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetVar( EPspReg dst_reg, const u32 * p_var )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( p_var, &reg_base, &base_offset );

	LW( dst_reg, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetFloatVar( EPspFloatReg dst_reg, const f32 * p_var )
{
	EPspReg		reg_base;
	s16			base_offset;
	GetBaseRegisterAndOffset( p_var, &reg_base, &base_offset );

	LWC1( dst_reg, reg_base, base_offset );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GetBaseRegisterAndOffset( const void * p_address, EPspReg * p_reg, s16 * p_offset )
{
	s32		base_pointer_offset( reinterpret_cast< const u8 * >( p_address ) - mpBasePointer );
	if( base_pointer_offset > SHRT_MIN && base_pointer_offset < SHRT_MAX )
	{
		*p_reg = mBaseRegister;
		*p_offset = base_pointer_offset;
	}
	else
	{
		u32		address( reinterpret_cast< u32 >( p_address ) );
		u16		hi_bits( (u16)( address >> 16 ) );
		u16		lo_bits( (u16)( address ) );

		//
		//	Move up
		//
		if(lo_bits >= 0x8000)
		{
			hi_bits++;
		}

		s32		long_offset( address - ((s32)hi_bits<<16) );

		DAEDALUS_ASSERT( long_offset >= SHRT_MIN && long_offset <= SHRT_MAX, "Offset is out of range!" );

		s16		offset( (s16)long_offset );

		DAEDALUS_ASSERT( ((s32)hi_bits<<16) + offset == (s32)address, "Incorrect address calculation" );

		LUI( PspReg_AT, hi_bits );
		*p_reg = PspReg_AT;
		*p_offset = offset;
	}
}

//*****************************************************************************
//	Generates instruction handler for the specified op code.
//	Returns a jump location if an exception handler is required
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateOpCode( u32 address, bool branch_delay_slot, OpCode op_code, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	bool	handled( false );
	bool	generate_exception_handler( false );
	bool	is_nop( op_code._u32 == 0 );

	if( is_nop )
	{
		if( branch_delay_slot )
		{
			SetVar( &gCPUState.Delay, NO_DELAY );
		}
		return CJumpLocation();
	}

	// See if this can be removed...
	if( branch_delay_slot )
	{
		SetVar( &gCPUState.Delay, EXEC_DELAY );
	}

	const EN64Reg	rs = EN64Reg( op_code.rs );
	const EN64Reg	rt = EN64Reg( op_code.rt );
	const EN64Reg	rd = EN64Reg( op_code.rd );
	const u32		sa = op_code.sa;
	const EN64Reg	base = EN64Reg( op_code.base );
	const u32		jump_target( (address&0xF0000000) | (op_code.target<<2) );
	const u32		branch_target( address + ( ((s32)(s16)op_code.immediate)<<2 ) + 4);
	const u32		ft = op_code.ft;

	CJumpLocation	exception_handler;
	// 
	//	Look for opcodes we can handle manually
	//
	switch( op_code.op )
	{

	case OP_JAL:		GenerateJAL( address, jump_target );	handled = true; generate_exception_handler = false; break;

	case OP_ADDI:		GenerateADDIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_ADDIU:		GenerateADDIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SLTI:		GenerateSLTI( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SLTIU:		GenerateSLTIU( rt, rs, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;

	case OP_ANDI:		GenerateANDI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_ORI:		GenerateORI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_XORI:		GenerateXORI( rt, rs, op_code.immediate );			handled = true; generate_exception_handler = false; break;
	case OP_LUI:		GenerateLUI( rt, s16( op_code.immediate ) );		handled = true; generate_exception_handler = false; break;

	case OP_LB:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLB( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_LBU:		SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLBU( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_LH:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLH( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_LHU:		SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLHU( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_LW:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLW( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;

	case OP_SB:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateSB( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SH:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateSH( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SW:			SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateSW( rt, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;

	case OP_LWC1:		SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateLWC1( ft, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;
	case OP_SWC1:		SetVar( &gCPUState.CurrentPC, address ); exception_handler = GenerateSWC1( ft, base, s16( op_code.immediate ) );	handled = true; generate_exception_handler = false; break;

	case OP_BEQ:		GenerateBEQ( rs, rt, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BNE:		GenerateBNE( rs, rt, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BLEZ:		GenerateBLEZ( rs, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;
	case OP_BGTZ:		GenerateBGTZ( rs, branch_target, p_branch, p_branch_jump ); handled = true; generate_exception_handler = false; break;

	case OP_SPECOP:
		switch( op_code.spec_op )
		{
		case SpecOp_SLL:	GenerateSLL( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRA:	GenerateSRA( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRL:	GenerateSRL( rd, rt, sa );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SLLV:	GenerateSLLV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRAV:	GenerateSRAV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SRLV:	GenerateSRLV( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
	
		case SpecOp_JR:		GenerateJR( rs, p_branch, p_branch_jump );	handled = true; generate_exception_handler = false; break;
		
		case SpecOp_MFLO:	GenerateMFLO( rd );			handled = true; generate_exception_handler = false; break;
		case SpecOp_MFHI:	GenerateMFHI( rd );			handled = true; generate_exception_handler = false; break;
		case SpecOp_MTLO:	GenerateMTLO( rd );			handled = true; generate_exception_handler = false; break;
		case SpecOp_MTHI:	GenerateMTHI( rd );			handled = true; generate_exception_handler = false; break;

		case SpecOp_MULT:	GenerateMULT( rs, rt );		handled = true; generate_exception_handler = false; break;
		case SpecOp_MULTU:	GenerateMULTU( rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_DIV:	GenerateDIV( rs, rt );		handled = true; generate_exception_handler = false; break;
		case SpecOp_DIVU:	GenerateDIVU( rs, rt );		handled = true; generate_exception_handler = false; break;

		case SpecOp_ADD:	GenerateADDU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_ADDU:	GenerateADDU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SUB:	GenerateSUBU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SUBU:	GenerateSUBU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		
		case SpecOp_AND:	GenerateAND( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_OR:		GenerateOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_XOR:	GenerateXOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_NOR:	GenerateNOR( rd, rs, rt );	handled = true; generate_exception_handler = false; break;

		case SpecOp_SLT:	GenerateSLT( rd, rs, rt );	handled = true; generate_exception_handler = false; break;
		case SpecOp_SLTU:	GenerateSLTU( rd, rs, rt );	handled = true; generate_exception_handler = false; break;

		default:
			break;
		}
		break;

	case OP_COPRO1:
		switch( op_code.cop1_op )
		{
		case Cop1Op_SInstr:
			switch( op_code.cop1_funct )
			{
			case Cop1OpFunc_ADD:	GenerateADD_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_SUB:	GenerateSUB_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_MUL:	GenerateMUL_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_DIV:	GenerateDIV_S( op_code.fd, op_code.fs, op_code.ft ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_SQRT:	GenerateSQRT_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_ABS:	GenerateABS_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_MOV:	GenerateMOV_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 
			case Cop1OpFunc_NEG:	GenerateNEG_S( op_code.fd, op_code.fs ); handled = true; generate_exception_handler = false; break; 

			}
			break;
		default:
			break;
		}
		break;
	}

	if( !handled )
	{
		//char msg[100];
		//SprintOpCodeInfo( msg, address, op_code );
		//printf( "Unhandled: %s\n", msg );

		//
		//	Default handling - call interpreting function
		//
		bool	need_pc( R4300_InstructionHandlerNeedsPC( op_code ) );

		if( !need_pc )
		{
			address = 0;
		}

		GenerateGenericR4300( address, op_code, R4300_GetInstructionHandler( op_code ) );

		generate_exception_handler = need_pc;
	}

	CCodeLabel		no_target( NULL );

	if( generate_exception_handler )
	{
		// Make sure all dirty registers are flushed. NB - we don't invalidate them
		// to avoid reloading the contents if no exception was thrown.

#ifdef STUFF_TO_DO_OPTIMISATION
		FlushAllRegisters( false );
		JAL( CCodeLabel( reinterpret_cast< const void * >( _ReturnFromDynaRecIfStuffToDo ) ), false );
		ORI( PspReg_A0, PspReg_R0, 0 );
#else
		FlushAllRegisters( false );
		exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, no_target );
#endif
	}

	// Check whether we want to invert the status of this branch
	if( p_branch != NULL )
	{
		if(!handled)
		{
			//
			// Check if the branch has been taken
			//
			if( p_branch->Direct )
			{
				if( p_branch->ConditionalBranchTaken )
				{
					*p_branch_jump = GenerateBranchIfNotEqual( &gCPUState.Delay, DO_DELAY, no_target );
				}
				else
				{
					*p_branch_jump = GenerateBranchIfEqual( &gCPUState.Delay, DO_DELAY, no_target );
				}
			}
			else
			{
				// XXXX eventually just exit here, and skip default exit code below
				if( p_branch->Eret )
				{
					*p_branch_jump = GenerateBranchAlways( no_target );
				}
				else
				{
					*p_branch_jump = GenerateBranchIfNotEqual( &gCPUState.TargetPC, p_branch->TargetAddress, no_target );
				}
			}
		}
	}
	else
	{
		if( branch_delay_slot )
		{
			SetVar( &gCPUState.Delay, NO_DELAY );
		}
	}

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateGenericR4300( u32 address, OpCode op_code, CPU_Instruction p_instruction )
{
	// Flush out the dirty registers, and mark them all as invalid
	// Theoretically we could just flush the registers referenced in the call
	FlushAllRegisters( true );

	if( address != 0 )
	{
		SetVar( &gCPUState.CurrentPC, address );
	}

	PspOpCode	op1;
	PspOpCode	op2;

	// Load the opcode in to A0
	GetLoadConstantOps( PspReg_A0, op_code._u32, &op1, &op2 );

	if( op2._u32 == 0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_instruction ) ), false );		// No branch delay
		AppendOp( op1 );
	}
	else
	{
		AppendOp( op1 );
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_instruction ) ), false );		// No branch delay
		AppendOp( op2 );
	}
}


//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::ExecuteSpeedHack( CCodeLabel speed_hack )
{
	JAL( speed_hack, true );
}

//*****************************************************************************
//
//*****************************************************************************
bool	CCodeGeneratorPSP::GenerateDirectLoad( EPspReg psp_dst, EN64Reg base, s16 offset, OpCodeValue load_op, u32 swizzle )
{
	if(mRegisterCache.IsKnownValue( base, 0 ))
	{
		u32		base_address( mRegisterCache.GetKnownValue( base, 0 ) );
		u32		address( (base_address + s32( offset )) ^ swizzle );

		s32		tableEntry = reinterpret_cast< s32 >( g_ReadAddressPointerLookupTable[address >> 18] ) + address;
		if(tableEntry >= 0)
		{
			void * p_memory( reinterpret_cast< void * >( tableEntry ) );

			//printf( "Loading from %s %08x + %04x (%08x) op %d\n", RegNames[ base ], base_address, u16(offset), address, load_op );
			
			EPspReg		reg_base;
			s16			base_offset;
			GetBaseRegisterAndOffset( p_memory, &reg_base, &base_offset );

			CAssemblyWriterPSP::LoadRegister( psp_dst, load_op, reg_base, base_offset );

			return true;
		}
		else
		{
			//printf( "Loading from %s %08x + %04x (%08x) - unhandled\n", RegNames[ base ], base_address, u16(offset), address );
		}
	}

	return false;
}

//*****************************************************************************
//
//*****************************************************************************
bool	CCodeGeneratorPSP::GenerateDirectStore( EN64Reg n64_src, EN64Reg base, s16 offset, OpCodeValue store_op, u32 swizzle )
{
	if(mRegisterCache.IsKnownValue( base, 0 ))
	{
		u32		base_address( mRegisterCache.GetKnownValue( base, 0 ) );
		u32		address( (base_address + s32( offset )) ^ swizzle );

		s32 tableEntry = reinterpret_cast< s32 >( g_WriteAddressPointerLookupTable[address >> 18] ) + address;
		if(tableEntry >= 0) // any way to tell MSC that tableEntry will typically be >= 0 ?
		{
			void * p_memory( reinterpret_cast< void * >( tableEntry ) );
			
			//printf( "Storing to %s %08x + %04x (%08x) op %d\n", RegNames[ base ], base_address, u16(offset), address, store_op );

			EPspReg		psp_src( GetRegisterAndLoadLo( n64_src, PspReg_T0 ) );

			EPspReg		reg_base;
			s16			base_offset;
			GetBaseRegisterAndOffset( p_memory, &reg_base, &base_offset );

			CAssemblyWriterPSP::StoreRegister( psp_src, store_op, reg_base, base_offset );

			return true;
		}
		else
		{
			//printf( "Storing to %s %08x + %04x (%08x) - unhandled\n", RegNames[ base ], base_address, u16(offset), address );
		}
	}

	return false;
}

//*****************************************************************************
//	Read 8 bits from memory (unsigned)
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLoad( EPspReg psp_dst, EN64Reg n64_base, s16 offset, OpCodeValue load_op, u32 swizzle, ReadMemoryFunction p_read_memory )
{
#ifdef ANALYSE_DIRECT_MEMORY_ACCESSES
	if(((gKnownLoads + gUnknownLoads)%500) == 0)
	{
		if(gKnownLoads+gUnknownLoads>0)
		{
			printf( "Known:Unknown = %d/%d (%f%%)\n", gKnownLoads, gUnknownLoads, float(gKnownLoads * 100) / float(gKnownLoads+gUnknownLoads) );
		}
	}
#endif

	//
	//	Check if the base pointer is a known value, in which case we can load directly.
	//
	if(GenerateDirectLoad( psp_dst, n64_base, offset, load_op, swizzle ))
	{
#ifdef ANALYSE_DIRECT_MEMORY_ACCESSES
		gKnownLoads++;
#endif
		return CJumpLocation();
	}
#ifdef ANALYSE_DIRECT_MEMORY_ACCESSES
	gUnknownLoads++;
#endif

	EPspReg		reg_base( GetRegisterAndLoadLo( n64_base, PspReg_A0 ) );
	EPspReg		reg_address;


	// We need to flush all the regular registers AND invalidate all the temp regs.
	// NB: When we can flush registers through some kind of handle (i.e. when the
	// exception handling branch is taken), this won't be necessary
	FlushAllRegisters( false );
	FlushAllTemporaryRegisters( true );

	if( offset != 0 )
	{
		ADDIU( PspReg_A0, reg_base, offset );
		reg_address = PspReg_A0;
	}
	else
	{
		reg_address = reg_base;
	}

	if( swizzle != 0 )
	{
		XORI( PspReg_A0, reg_address, swizzle );
		reg_address = PspReg_A0;
	}

	SRL( PspReg_A1, reg_address, 0x12 );		// >> 18
	SLL( PspReg_V0, PspReg_A1, 0x2 );			// *4
	ADDU( PspReg_V0, PspReg_V0, PspReg_S7 );	// S7 is g_ReadAddressPointerLookupTable
	LW( PspReg_V0, PspReg_V0, 0 );				// Get table entry
	ADDU( PspReg_V0, reg_address, PspReg_V0 );	// Add offset
	CJumpLocation branch( BGEZL( PspReg_V0, CCodeLabel(NULL), false ) );
	CAssemblyWriterPSP::LoadRegister( psp_dst, load_op, PspReg_V0, 0 );

	if( reg_address != PspReg_A0 )
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_read_memory ) ), false );
		OR( PspReg_A0, reg_address, PspReg_R0 );
	}
	else
	{
		JAL( CCodeLabel( reinterpret_cast< const void * >( p_read_memory ) ), true );
	}
	// Copy return value to the destination register if we're not using V0
	if( psp_dst != PspReg_V0 )
	{
		OR( psp_dst, PspReg_V0, PspReg_R0 );
	}

	// We only need to generate an exception handler on calls to p_read_memory
	CJumpLocation	exception_handler;
#ifndef STUFF_TO_DO_OPTIMISATION
	exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, CCodeLabel(NULL) );
#endif

	// Registers flushed above
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateJAL( u32 address, u32 target )
{
	//gGPR[REG_ra]._s64 = (s64)(s32)(gCPUState.CurrentPC + 8);		// Store return address
	//u32	new_pc( (gCPUState.CurrentPC & 0xF0000000) | (op_code.target<<2) );
	//CPU_TakeBranch( new_pc, CPU_BRANCH_DIRECT );

	EPspReg	reg_lo( GetRegisterNoLoadLo( N64Reg_RA, PspReg_T0 ) );
	LoadConstant( reg_lo, address + 8 );
	UpdateRegister( N64Reg_RA, reg_lo, URO_HI_SIGN_EXTEND, PspReg_T0 );

	// Necessary?
	//SetVar( &gCPUState.TargetPC, target );
	//SetVar( &gCPUState.Delay, DO_DELAY );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateJR( EN64Reg rs, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	//u32	new_pc( gGPR[ op_code.rs ]._u32[0] );
	//CPU_TakeBranch( new_pc, CPU_BRANCH_INDIRECT );

	EPspReg reg_lo( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	// Necessary? Could just directly compare reg_lo and constant p_branch->TargetAddress??
	//SetVar( &gCPUState.TargetPC, reg_lo );
	//SetVar( &gCPUState.Delay, DO_DELAY );
	//*p_branch_jump = GenerateBranchIfNotEqual( &gCPUState.TargetPC, p_branch->TargetAddress, CCodeLabel() );

	SetVar( &gCPUState.TargetPC, reg_lo );
	//SetVar( &gCPUState.Delay, DO_DELAY );
	*p_branch_jump = GenerateBranchIfNotEqual( reg_lo, p_branch->TargetAddress, CCodeLabel() );


}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFLO( EN64Reg rd )
{
	//gGPR[ op_code.rd ]._u64 = g_qwMultLO._u64;

	EPspReg	reg_lo( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	GetVar( reg_lo, &g_qwMultLO._u32[0] );
	StoreRegisterLo( rd, reg_lo );

	EPspReg	reg_hi( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	GetVar( reg_hi, &g_qwMultLO._u32[1] );
	StoreRegisterHi( rd, reg_hi );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFHI( EN64Reg rd )
{
	//gGPR[ op_code.rd ]._u64 = g_qwMultHI._u64;

	EPspReg	reg_lo( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	GetVar( reg_lo, &g_qwMultHI._u32[0] );
	StoreRegisterLo( rd, reg_lo );

	EPspReg	reg_hi( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	GetVar( reg_hi, &g_qwMultHI._u32[1] );
	StoreRegisterHi( rd, reg_hi );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMTLO( EN64Reg rs )
{
	//g_qwMultLO._u64 = gGPR[ op_code.rs ]._u64;

	EPspReg	reg_lo( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	SetVar( &g_qwMultLO._u32[0], reg_lo );

	EPspReg	reg_hi( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	SetVar( &g_qwMultLO._u32[1], reg_hi );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMTHI( EN64Reg rs )
{
	//g_qwMultHI._u64 = gGPR[ op_code.rs ]._u64;

	EPspReg	reg_lo( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	SetVar( &g_qwMultHI._u32[0], reg_lo );

	EPspReg	reg_hi( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	SetVar( &g_qwMultHI._u32[1], reg_hi );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMULT( EN64Reg rs, EN64Reg rt )
{
	//s64 dwResult = (s64)gGPR[ op_code.rs ]._s32[0] * (s64)gGPR[ op_code.rt ]._s32[0];
	//g_qwMultLO = (s64)(s32)(dwResult);
	//g_qwMultHI = (s64)(s32)(dwResult >> 32);

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	MULT( reg_lo_a, reg_lo_b );

	MFLO( PspReg_T0 );
	MFHI( PspReg_T1 );

	SetVar( &g_qwMultLO._u32[0], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[0], PspReg_T1 );

	SRA( PspReg_T0, PspReg_T0, 0x1f );		// Sign extend
	SRA( PspReg_T1, PspReg_T1, 0x1f );		// Sign extend

	SetVar( &g_qwMultLO._u32[1], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[1], PspReg_T1 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMULTU( EN64Reg rs, EN64Reg rt )
{
	//u64 dwResult = (u64)gGPR[ op_code.rs ]._u32[0] * (u64)gGPR[ op_code.rt ]._u32[0];
	//g_qwMultLO = (s64)(s32)(dwResult);
	//g_qwMultHI = (s64)(s32)(dwResult >> 32);

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	MULTU( reg_lo_a, reg_lo_b );

	MFLO( PspReg_T0 );
	MFHI( PspReg_T1 );

	SetVar( &g_qwMultLO._u32[0], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[0], PspReg_T1 );

	SRA( PspReg_T0, PspReg_T0, 0x1f );		// Sign extend
	SRA( PspReg_T1, PspReg_T1, 0x1f );		// Sign extend

	SetVar( &g_qwMultLO._u32[1], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[1], PspReg_T1 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV( EN64Reg rs, EN64Reg rt )
{
	//s32 nDividend = gGPR[ op_code.rs ]._s32[0];
	//s32 nDivisor  = gGPR[ op_code.rt ]._s32[0];

	//if (nDivisor)
	//{
	//	g_qwMultLO._u64 = (s64)(s32)(nDividend / nDivisor);
	//	g_qwMultHI._u64 = (s64)(s32)(nDividend % nDivisor);
	//}

	EPspReg	reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	CJumpLocation	branch( BEQ( reg_lo_rt, PspReg_R0, CCodeLabel(NULL), true ) );		// Can use branch delay for something?

	DIV( reg_lo_rs, reg_lo_rt );

	MFLO( PspReg_T0 );
	MFHI( PspReg_T1 );

	SetVar( &g_qwMultLO._u32[0], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[0], PspReg_T1 );

	SRA( PspReg_T0, PspReg_T0, 0x1f );		// Sign extend
	SRA( PspReg_T1, PspReg_T1, 0x1f );		// Sign extend

	SetVar( &g_qwMultLO._u32[1], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[1], PspReg_T1 );

	// Branch here - really should trigger exception!
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIVU( EN64Reg rs, EN64Reg rt )
{
	//u32 dwDividend = gGPR[ op_code.rs ]._u32[0];
	//u32 dwDivisor  = gGPR[ op_code.rt ]._u32[0];

	//if (dwDivisor) {
	//	g_qwMultLO._u64 = (s64)(s32)(dwDividend / dwDivisor);
	//	g_qwMultHI._u64 = (s64)(s32)(dwDividend % dwDivisor);
	//}

	EPspReg	reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	CJumpLocation	branch( BEQ( reg_lo_rt, PspReg_R0, CCodeLabel(NULL), true ) );		// Can use branch delay for something?

	DIVU( reg_lo_rs, reg_lo_rt );

	MFLO( PspReg_T0 );
	MFHI( PspReg_T1 );

	SetVar( &g_qwMultLO._u32[0], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[0], PspReg_T1 );

	SRA( PspReg_T0, PspReg_T0, 0x1f );		// Sign extend
	SRA( PspReg_T1, PspReg_T1, 0x1f );		// Sign extend

	SetVar( &g_qwMultLO._u32[1], PspReg_T0 );
	SetVar( &g_qwMultHI._u32[1], PspReg_T1 );

	// Branch here - really should trigger exception!
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rs ]._s32[0] + gGPR[ op_code.rt ]._s32[0] );
	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	ADDU( reg_lo_d, reg_lo_a, reg_lo_b );
	UpdateRegister( rd, reg_lo_d, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUBU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rs ]._s32[0] - gGPR[ op_code.rt ]._s32[0] );
	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	SUBU( reg_lo_d, reg_lo_a, reg_lo_b );
	UpdateRegister( rd, reg_lo_d, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateAND( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 & gGPR[ op_code.rt ]._u64;

	if( rs == N64Reg_R0 || rt == N64Reg_R0 )
	{
		SetRegisterLo( rd, 0, PspReg_T0 );
		SetRegisterHi( rd, 0, PspReg_T0 );
	}
	else
	{
		// XXXX or into dest register
		EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
		EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
		AND( reg_lo_d, reg_lo_a, reg_lo_b );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
		EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
		AND( reg_hi_d, reg_hi_a, reg_hi_b );
		StoreRegisterHi( rd, reg_hi_d );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 | gGPR[ op_code.rt ]._u64;

	if( rs == N64Reg_R0 && rt == N64Reg_R0 )
	{
		SetRegisterLo( rd, 0, PspReg_T0 );
		SetRegisterHi( rd, 0, PspReg_T0 );
	}
	else if( rs == N64Reg_R0 )
	{
		// This case rarely seems to happen...
		// As RS is zero, the OR is just a copy of RT to RD.
		// Try to avoid loading into a temp register if the dest is cached
		EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		LoadRegisterLo( reg_lo_d, rt );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		LoadRegisterHi( reg_hi_d, rt );
		StoreRegisterHi( rd, reg_hi_d );
	}
	else if( rt == N64Reg_R0 )
	{
		// As RT is zero, the OR is just a copy of RS to RD.
		// Try to avoid loading into a temp register if the dest is cached
		EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		LoadRegisterLo( reg_lo_d, rs );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		LoadRegisterHi( reg_hi_d, rs );
		StoreRegisterHi( rd, reg_hi_d );
	}
	else
	{
		EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
		EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
		EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
		OR( reg_lo_d, reg_lo_a, reg_lo_b );
		StoreRegisterLo( rd, reg_lo_d );

		EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
		EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
		EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
		OR( reg_hi_d, reg_hi_a, reg_hi_b );
		StoreRegisterHi( rd, reg_hi_d );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = gGPR[ op_code.rs ]._u64 ^ gGPR[ op_code.rt ]._u64;

	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	XOR( reg_lo_d, reg_lo_a, reg_lo_b );
	StoreRegisterLo( rd, reg_lo_d );

	EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
	XOR( reg_hi_d, reg_hi_a, reg_hi_b );
	StoreRegisterHi( rd, reg_hi_d );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNOR( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._u64 = ~(gGPR[ op_code.rs ]._u64 | gGPR[ op_code.rt ]._u64);

	EPspReg	reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );
	NOR( reg_lo_d, reg_lo_a, reg_lo_b );
	StoreRegisterLo( rd, reg_lo_d );

	EPspReg	reg_hi_d( GetRegisterNoLoadHi( rd, PspReg_T0 ) );
	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );
	NOR( reg_hi_d, reg_hi_a, reg_hi_b );
	StoreRegisterHi( rd, reg_hi_d );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLT( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the registers regardless of whether we take it or not. We pull in the lo halves of 
	// the registers here so that they're Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );
	PrepareCachedRegisterLo( rt );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );

	if(rd == rs || rd == rt)
	{
		reg_lo_d = PspReg_T0;
	}

	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );

	CJumpLocation	branch( BNE( reg_hi_a, reg_hi_b, CCodeLabel(NULL), false ) );
	SLT( reg_lo_d, reg_hi_a, reg_hi_b );		// In branch delay slot

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	SLT( reg_lo_d, reg_lo_a, reg_lo_b );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rd, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTU( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the registers regardless of whether we take it or not. We pull in the lo halves of 
	// the registers here so that they're Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );
	PrepareCachedRegisterLo( rt );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rd, PspReg_T0 ) );

	if(rd == rs || rd == rt)
	{
		reg_lo_d = PspReg_T0;
	}

	EPspReg	reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	EPspReg	reg_hi_b( GetRegisterAndLoadHi( rt, PspReg_T1 ) );

	CJumpLocation	branch( BNE( reg_hi_a, reg_hi_b, CCodeLabel(NULL), false ) );
	SLTU( reg_lo_d, reg_hi_a, reg_hi_b );		// In branch delay slot

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half

	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg	reg_lo_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	SLTU( reg_lo_d, reg_lo_a, reg_lo_b );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rd, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADDIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	//gGPR[op_code.rt]._s64 = (s64)(s32)(gGPR[op_code.rs]._s32[0] + (s32)(s16)op_code.immediate);
	bool	known = false;
	s32		known_value = 0;

	if(mRegisterCache.IsKnownValue( rs, 0 ))
	{
		known_value = mRegisterCache.GetKnownValue( rs, 0 ) + (s32)immediate;
		known = true;
	}

	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	ADDIU( dst_reg, src_reg, immediate );
	UpdateRegister( rt, dst_reg, URO_HI_SIGN_EXTEND, PspReg_T0 );

	if(known)
	{
		mRegisterCache.SetKnownValue( rt, 0, known_value );
		mRegisterCache.SetKnownValue( rt, 1, known_value >= 0 ? 0 : 0xffffffff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateANDI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 & (u64)(u16)op_code.immediate;
	bool	known = false;
	s32		known_value = 0;

	if(mRegisterCache.IsKnownValue( rs, 0 ))
	{
		known_value = mRegisterCache.GetKnownValue( rs, 0 ) & (u32)immediate;
		known = true;
	}

	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	ANDI( dst_reg, src_reg, immediate );
	UpdateRegister( rt, dst_reg, URO_HI_CLEAR, PspReg_T0 );

	if(known)
	{
		mRegisterCache.SetKnownValue( rt, 0, known_value );
		mRegisterCache.SetKnownValue( rt, 1, known_value >= 0 ? 0 : 0xffffffff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 | (u64)(u16)op_code.immediate;

	bool	known = false;
	s32		known_value = 0;

	if(mRegisterCache.IsKnownValue( rs, 0 ))
	{
		known_value = mRegisterCache.GetKnownValue( rs, 0 ) | (u32)immediate;
		known = true;
	}

	if( rs == N64Reg_R0 )
	{
		// This may evaluate to the same code as below, but it means we can keep track of the known values and propagate
		SetRegisterLo( rt, immediate, PspReg_T0 );
		SetRegisterHi( rt, 0, PspReg_T0 );
	}
	else
	{
		EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
		EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
		ORI( dst_reg, src_reg, immediate );
		StoreRegisterLo( rt, dst_reg );

		// If the source/dest regs are different we need to copy the high bits across
		if(rt != rs)
		{
			EPspReg dst_reg_hi( GetRegisterNoLoadHi( rt, PspReg_T0 ) );
			LoadRegisterHi( dst_reg_hi, rs );
			StoreRegisterHi( rt, dst_reg_hi );
		}
	}

	if(known)
	{
		mRegisterCache.SetKnownValue( rt, 0, known_value );
		mRegisterCache.SetKnownValue( rt, 1, known_value >= 0 ? 0 : 0xffffffff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateXORI( EN64Reg rt, EN64Reg rs, u16 immediate )
{
	//gGPR[op_code.rt]._u64 = gGPR[op_code.rs]._u64 ^ (u64)(u16)op_code.immediate;

	bool	known = false;
	s32		known_value = 0;

	if(mRegisterCache.IsKnownValue( rs, 0 ))
	{
		known_value = mRegisterCache.GetKnownValue( rs, 0 ) ^ (u32)immediate;
		known = true;
	}

	EPspReg dst_reg( GetRegisterNoLoadLo( rt, PspReg_T0 ) );
	EPspReg	src_reg( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	XORI( dst_reg, src_reg, immediate );
	StoreRegisterLo( rt, dst_reg );

	// If the source/dest regs are different we need to copy the high bits across
	// (if they are the same, we're xoring 0 to the top half which is essentially a NOP)
	if(rt != rs)
	{
		EPspReg dst_reg_hi( GetRegisterNoLoadHi( rt, PspReg_T0 ) );
		LoadRegisterHi( dst_reg_hi, rs );
		StoreRegisterHi( rt, dst_reg_hi );
	}

	if(known)
	{
		mRegisterCache.SetKnownValue( rt, 0, known_value );
		mRegisterCache.SetKnownValue( rt, 1, known_value >= 0 ? 0 : 0xffffffff );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateLUI( EN64Reg rt, s16 immediate )
{
	//gGPR[op_code.rt]._s64 = (s64)(s32)((s32)(s16)op_code.immediate<<16);

	// Load into cached reg
	LUI( PspReg_T0, immediate );

	if (immediate >= 0)
	{
		UpdateRegister( rt, PspReg_T0, URO_HI_CLEAR, PspReg_T0 );
	}
	else
	{
		UpdateRegister( rt, PspReg_T0, URO_HI_SIGN_EXTEND, PspReg_T0 );
	}

	s32	known_value( s32( immediate ) << 16 );

	mRegisterCache.SetKnownValue( rt, 0, known_value );
	mRegisterCache.SetKnownValue( rt, 1, known_value >= 0 ? 0 : 0xffffffff );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTI( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the register regardless of whether we take it or not. We pull in the lo halves of 
	// the register here so that it's Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rt, PspReg_T0 ) );

	if(rt == rs)
	{
		reg_lo_d = PspReg_T0;
	}

	CJumpLocation	branch;

	EPspReg		reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	if( immediate >= 0 )
	{
		// Positive data - we can avoid a contant load here
		branch = BNE( reg_hi_a, PspReg_R0, CCodeLabel(NULL), false );
		SLTI( reg_lo_d, reg_hi_a, 0x0000 );		// In branch delay slot
	}
	else
	{
		// Negative data
		LoadConstant( PspReg_T1, -1 );
		branch = BNE( reg_hi_a, PspReg_T1, CCodeLabel(NULL), false );
		SLTI( reg_lo_d, reg_hi_a, 0xffff );		// In branch delay slot
	}

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	SLTI( reg_lo_d, reg_lo_a, immediate );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rt, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLTIU( EN64Reg rt, EN64Reg rs, s16 immediate )
{
	// Because we have a branch here, we need to make sure that we have a consistant view
	// of the register regardless of whether we take it or not. We pull in the lo halves of 
	// the register here so that it's Valid regardless of whether we take the branch or not
	PrepareCachedRegisterLo( rs );

	// If possible, we write directly into the destination register. We have to be careful though -
	// if the destination register is the same as either of the source registers we have to use
	// a temporary instead, to avoid overwriting the contents.
	EPspReg reg_lo_d( GetRegisterNoLoadLo( rt, PspReg_T0 ) );

	if(rt == rs)
	{
		reg_lo_d = PspReg_T0;
	}

	CJumpLocation	branch;

	EPspReg		reg_hi_a( GetRegisterAndLoadHi( rs, PspReg_T0 ) );
	if( immediate >= 0 )
	{
		// Positive data - we can avoid a contant load here
		branch = BNE( reg_hi_a, PspReg_R0, CCodeLabel(NULL), false );
		SLTIU( reg_lo_d, reg_hi_a, 0x0000 );		// In branch delay slot
	}
	else
	{
		// Negative data
		LoadConstant( PspReg_T1, -1 );
		branch = BNE( reg_hi_a, PspReg_T1, CCodeLabel(NULL), false );
		SLTIU( reg_lo_d, reg_hi_a, 0xffff );		// In branch delay slot
	}

	// If the branch was not taken, it means that the high part of the registers was equal, so compare bottom half
	EPspReg	reg_lo_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	SLTIU( reg_lo_d, reg_lo_a, immediate );

	// Patch up the branch
	PatchJumpLong( branch, GetAssemblyBuffer()->GetLabel() );

	UpdateRegister( rt, reg_lo_d, URO_HI_CLEAR, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( (gGPR[ op_code.rt ]._u32[0] << op_code.sa) & 0xFFFFFFFF );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SLL( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRL( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._u32[0] >> op_code.sa );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRL( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRA( EN64Reg rd, EN64Reg rt, u32 sa )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._s32[0] >> op_code.sa );

	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRA( reg_lo_rd, reg_lo_rt, sa );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSLLV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( (gGPR[ op_code.rt ]._u32[0] << ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) ) & 0xFFFFFFFF );

	// PSP sllv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SLLV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRLV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._u32[0] >> ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) );

	// PSP srlv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRLV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSRAV( EN64Reg rd, EN64Reg rs, EN64Reg rt )
{
	//gGPR[ op_code.rd ]._s64 = (s64)(s32)( gGPR[ op_code.rt ]._s32[0] >> ( gGPR[ op_code.rs ]._u32[0] & 0x1F ) );

	// PSP srlv does exactly the same op as n64- no need for masking
	EPspReg reg_lo_rd( GetRegisterNoLoadLo( rd, PspReg_T0 ) );
	EPspReg reg_lo_rs( GetRegisterAndLoadLo( rs, PspReg_T1 ) );
	EPspReg	reg_lo_rt( GetRegisterAndLoadLo( rt, PspReg_T0 ) );

	SRAV( reg_lo_rd, reg_lo_rs, reg_lo_rt );
	UpdateRegister( rd, reg_lo_rd, URO_HI_SIGN_EXTEND, PspReg_T0 );
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLB( EN64Reg rt, EN64Reg base, s32 offset )
{
	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LB, 3, ReadBitsDirect_s8 );	// NB this fills the whole of reg_dst

	UpdateRegister( rt, reg_dst, URO_HI_SIGN_EXTEND, PspReg_T0 );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLBU( EN64Reg rt, EN64Reg base, s32 offset )
{
	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LBU, 3, ReadBitsDirect_u8 );	// NB this fills the whole of reg_dst

	UpdateRegister( rt, reg_dst, URO_HI_CLEAR, PspReg_T0 );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLH( EN64Reg rt, EN64Reg base, s32 offset )
{
	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LH, 2, ReadBitsDirect_s16 );	// NB this fills the whole of reg_dst

	UpdateRegister( rt, reg_dst, URO_HI_SIGN_EXTEND, PspReg_T0 );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLHU( EN64Reg rt, EN64Reg base, s32 offset )
{
	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LHU, 2, ReadBitsDirect_u16 );	// NB this fills the whole of reg_dst

	UpdateRegister( rt, reg_dst, URO_HI_CLEAR, PspReg_T0 );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLW( EN64Reg rt, EN64Reg base, s16 offset )
{
	EPspReg	reg_dst( GetRegisterNoLoadLo( rt, PspReg_V0 ) );	// Use V0 to avoid copying return value if reg is not cached

	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LW, 0, ReadBitsDirect_u32 );

	UpdateRegister( rt, reg_dst, URO_HI_SIGN_EXTEND, PspReg_T0 );
	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateLWC1( u32 ft, EN64Reg base, s32 offset )
{
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EPspFloatReg	psp_ft( GetFloatRegisterNoLoad( n64_ft ) );

	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//value = Read32Bits(address);
	//g_qwCPR[1][op_code.ft]._s32[ 0 ] = value;

	// TODO: Actually perform LWC1 here
	EPspReg	reg_dst( PspReg_V0 );				// GenerateLoad32 is slightly more efficient when using V0
	CJumpLocation	exception_handler;
	exception_handler = GenerateLoad( reg_dst, base, offset, OP_LW, 0, ReadBitsDirect_u32 );

	//SetVar( &g_qwCPR[1][ ft ]._u32[ 0 ], reg_dst );
	MTC1( psp_ft, reg_dst );
	UpdateFloatRegister( n64_ft );

	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateSB( EN64Reg rt, EN64Reg base, s32 offset )
{
	if(GenerateDirectStore( rt, base, offset, OP_SB, 3 ))
	{
		return CJumpLocation();
	}

	LoadRegisterLo( PspReg_A0, base );
	LoadRegisterLo( PspReg_A1, rt );

	// We need to flush all the regular registers AND invalidate all the temp regs.
	// NB: When we can flush registers through some kind of handle (i.e. when the
	// exception handling branch is taken), this won't be necessary
	FlushAllRegisters( false );
	FlushAllTemporaryRegisters( true );
	JAL( CCodeLabel( reinterpret_cast< const void * >( WriteBitsDirect_u8 ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );

	CJumpLocation	exception_handler;
#ifndef STUFF_TO_DO_OPTIMISATION
	exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, CCodeLabel(NULL) );
#endif
	return exception_handler;

}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateSH( EN64Reg rt, EN64Reg base, s32 offset )
{
	if(GenerateDirectStore( rt, base, offset, OP_SH, 2 ))
	{
		return CJumpLocation();
	}

	LoadRegisterLo( PspReg_A0, base );
	LoadRegisterLo( PspReg_A1, rt );

	// We need to flush all the regular registers AND invalidate all the temp regs.
	// NB: When we can flush registers through some kind of handle (i.e. when the
	// exception handling branch is taken), this won't be necessary
	FlushAllRegisters( false );
	FlushAllTemporaryRegisters( true );
	JAL( CCodeLabel( reinterpret_cast< const void * >( WriteBitsDirect_u16 ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );

	CJumpLocation	exception_handler;
#ifndef STUFF_TO_DO_OPTIMISATION
	exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, CCodeLabel(NULL) );
#endif
	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateSW( EN64Reg rt, EN64Reg base, s32 offset )
{
	if(GenerateDirectStore( rt, base, offset, OP_SW, 0 ))
	{
		return CJumpLocation();
	}

	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//Write32Bits(address, gGPR[op_code.rt]._u32[0]);

	LoadRegisterLo( PspReg_A0, base );
	LoadRegisterLo( PspReg_A1, rt );

	// We need to flush all the regular registers AND invalidate all the temp regs.
	// NB: When we can flush registers through some kind of handle (i.e. when the
	// exception handling branch is taken), this won't be necessary
	FlushAllRegisters( false );
	FlushAllTemporaryRegisters( true );
	JAL( CCodeLabel( reinterpret_cast< const void * >( WriteBitsDirect_u32 ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );

	CJumpLocation	exception_handler;
#ifndef STUFF_TO_DO_OPTIMISATION
	exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, CCodeLabel(NULL) );
#endif
	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
CJumpLocation	CCodeGeneratorPSP::GenerateSWC1( u32 ft, EN64Reg base, s32 offset )
{
	//u32 address = (u32)( gGPR[op_code.base]._s32[0] + (s32)(s16)op_code.immediate );
	//Write32Bits(address, g_qwCPR[1][op_code.ft]._s32[ 0 ] );

	LoadRegisterLo( PspReg_A0, base );
	//GetVar( PspReg_A1, &g_qwCPR[1][ft]._u32[ 0 ] );

	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );
	MFC1( PspReg_A1, psp_ft );

	// We need to flush all the regular registers AND invalidate all the temp regs.
	// NB: When we can flush registers through some kind of handle (i.e. when the
	// exception handling branch is taken), this won't be necessary
	FlushAllRegisters( false );
	FlushAllTemporaryRegisters( true );
	JAL( CCodeLabel( reinterpret_cast< const void * >( WriteBitsDirect_u32 ) ), false );		// Don't set branch delay slot
	ADDIU( PspReg_A0, PspReg_A0, offset );

	CJumpLocation	exception_handler;
#ifndef STUFF_TO_DO_OPTIMISATION
	exception_handler = GenerateBranchIfSet( &gCPUState.StuffToDo, CCodeLabel(NULL) );
#endif
	return exception_handler;
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMFC1( EN64Reg rt, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMTC1( u32 fs, EN64Reg rt )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCFC1( EN64Reg rt, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCTC1( u32 fs, EN64Reg rt )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBEQ( EN64Reg rs, EN64Reg rt, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BEQ?" );

	// One or other of these may be r0 - we don't really care for optimisation purposes though
	// as ultimately the register load regs factored out
	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg		reg_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BNE( reg_a, reg_b, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BEQ( reg_a, reg_b, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBNE( EN64Reg rs, EN64Reg rt, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BNE?" );

	// One or other of these may be r0 - we don't really care for optimisation purposes though
	// as ultimately the register load regs factored out
	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );
	EPspReg		reg_b( GetRegisterAndLoadLo( rt, PspReg_T1 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BEQ( reg_a, reg_b, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BNE( reg_a, reg_b, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBLEZ( EN64Reg rs, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BLEZ?" );

	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BGTZ( reg_a, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BLEZ( reg_a, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateBGTZ( EN64Reg rs, u32 target, const SBranchDetails * p_branch, CJumpLocation * p_branch_jump )
{
	DAEDALUS_ASSERT( p_branch != NULL, "No branch details?" );
	DAEDALUS_ASSERT( p_branch->Direct, "Indirect branch for BGTZ?" );

	EPspReg		reg_a( GetRegisterAndLoadLo( rs, PspReg_T0 ) );

	// XXXX This may actually need to be a 64 bit compare, but this is what R4300.cpp does

	if( p_branch->ConditionalBranchTaken )
	{
		// Flip the sign of the test - 
		*p_branch_jump = BLEZ( reg_a, CCodeLabel(NULL), true );
	}
	else
	{
		*p_branch_jump = BGTZ( reg_a, CCodeLabel(NULL), true );
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	ADD_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	SUB_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	MUL_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_S( u32 fd, u32 fs, u32 ft )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_ft = EN64FloatReg( ft );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );
	EPspFloatReg	psp_ft( GetFloatRegisterAndLoad( n64_ft ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	DIV_S( psp_fd, psp_fs, psp_ft );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	SQRT_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateABS_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	ABS_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	MOV_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_S( u32 fd, u32 fs )
{
	EN64FloatReg	n64_fs = EN64FloatReg( fs );
	EN64FloatReg	n64_fd = EN64FloatReg( fd );

	EPspFloatReg	psp_fd( GetFloatRegisterNoLoad( n64_fd ) );
	EPspFloatReg	psp_fs( GetFloatRegisterAndLoad( n64_fs ) );

	//SET_ROUND_MODE( gRoundingMode );		//XXXX Is this needed?

	NEG_S( psp_fd, psp_fs );

	UpdateFloatRegister( n64_fd );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_S( u32 fs, u32 ft, u16 flag )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateADD_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSUB_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMUL_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateDIV_D( u32 fd, u32 fs, u32 ft )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateSQRT_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateMOV_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateNEG_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCompare_D( u32 fs, u32 ft, u16 flag )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_S( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_W_S( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_D_W( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_W( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

//*****************************************************************************
//
//*****************************************************************************
void	CCodeGeneratorPSP::GenerateCVT_S_D( u32 fd, u32 fs )
{
	NOT_IMPLEMENTED( __FUNCTION__ );
}

