/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef PSPCOLOUR_H__
#define PSPCOLOUR_H__

#include "DaedMathUtil.h"

inline u32 MAKE_COLOUR_U32( s32 r, s32 g, s32 b )
{
	return (0xff << 24) | ((b&255)<<16) | ((g&255)<<8) | (r&255);
}

inline u32 MAKE_COLOUR_U32( s32 r, s32 g, s32 b, s32 a )
{
	return ((a&255) << 24) | ((b&255)<<16) | ((g&255)<<8) | (r&255);
}

inline u8	COL_GETA( u32 abgr )
{
	return (abgr >> 24) & 255;
}
inline u8	COL_GETB( u32 abgr )
{
	return (abgr >> 16) & 255;
}
inline u8	COL_GETG( u32 abgr )
{
	return (abgr >>  8) & 255;
}
inline u8	COL_GETR( u32 abgr )
{
	return (abgr      ) & 255;
}

inline float COL_GETA_F( u32 abgr ) { return COL_GETA( abgr ) / 255.0f; }
inline float COL_GETB_F( u32 abgr ) { return COL_GETB( abgr ) / 255.0f; }
inline float COL_GETG_F( u32 abgr ) { return COL_GETG( abgr ) / 255.0f; }
inline float COL_GETR_F( u32 abgr ) { return COL_GETR( abgr ) / 255.0f; }


inline u32 MAKE_COLOUR_U32_REPLICATE_ALPHA( u32 col )
{
	u32		a( COL_GETA( col ) );

	return MAKE_COLOUR_U32( a, a, a, a );
}

// 1-r, 1-g, 1-b, a
inline u32 MAKE_COLOUR_U32_RGB_1MC( u32 col )
{
	u32		a( (u32( COL_GETA( col ) ) ) );
	u32		b( 255 - (u32( COL_GETB( col ) ) ) );
	u32		g( 255 - (u32( COL_GETG( col ) ) ) );
	u32		r( 255 - (u32( COL_GETR( col ) ) ) );

	return MAKE_COLOUR_U32( r, g, b, a );
}

// Modulate RGB, select cola A
inline u32	MODULATE_COLOUR_RGB( u32 cola, u32 colb )
{
	u32		a( (u32( COL_GETA( cola ) ) ) );
	u32		b( (u32( COL_GETB( cola ) ) * u32( COL_GETB( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		g( (u32( COL_GETG( cola ) ) * u32( COL_GETG( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		r( (u32( COL_GETR( cola ) ) * u32( COL_GETR( colb ) )) >> 8 );		// >> 8 to return to 0..255

	return MAKE_COLOUR_U32( r, g, b, a );
}

// Select cola RGB, modulate A
inline u32	MODULATE_COLOUR_A( u32 cola, u32 colb )
{
	u32		a( (u32( COL_GETA( cola ) ) * u32( COL_GETA( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		b( (u32( COL_GETB( cola ) ) ) );
	u32		g( (u32( COL_GETG( cola ) ) ) );
	u32		r( (u32( COL_GETR( cola ) ) ) );

	return MAKE_COLOUR_U32( r, g, b, a );
}


// Modulate RGBA
inline u32	MODULATE_COLOUR_RGBA( u32 cola, u32 colb )
{
	u32		a( (u32( COL_GETA( cola ) ) * u32( COL_GETA( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		b( (u32( COL_GETB( cola ) ) * u32( COL_GETB( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		g( (u32( COL_GETG( cola ) ) * u32( COL_GETG( colb ) )) >> 8 );		// >> 8 to return to 0..255
	u32		r( (u32( COL_GETR( cola ) ) * u32( COL_GETR( colb ) )) >> 8 );		// >> 8 to return to 0..255

	return MAKE_COLOUR_U32( r, g, b, a );
}


// Add RGBA
inline u32	ADD_COLOUR_RGBA( u32 cola, u32 colb )
{
	u32		a( Clamp< u32 >( (u32( COL_GETA( cola ) ) + u32( COL_GETA( colb ) )), 0, 255 ) );
	u32		b( Clamp< u32 >( (u32( COL_GETB( cola ) ) + u32( COL_GETB( colb ) )), 0, 255 ) );
	u32		g( Clamp< u32 >( (u32( COL_GETG( cola ) ) + u32( COL_GETG( colb ) )), 0, 255 ) );
	u32		r( Clamp< u32 >( (u32( COL_GETR( cola ) ) + u32( COL_GETR( colb ) )), 0, 255 ) );

	return MAKE_COLOUR_U32( r, g, b, a );
}

inline u8	Interpolate( u8 a, u8 b, float factor )
{
	return u8(float(a) + (float(b) - float(a)) * factor);
}

inline u32 COL32_BLEND( u32 cola, u32 colb, u32 factor )
{
	float	factor_a( COL_GETA_F( factor ) );
	float	factor_b( COL_GETB_F( factor ) );
	float	factor_g( COL_GETG_F( factor ) );
	float	factor_r( COL_GETR_F( factor ) );

	s32		a( Interpolate( COL_GETA( cola ), COL_GETA( colb ), factor_a ) );
	s32		b( Interpolate( COL_GETB( cola ), COL_GETB( colb ), factor_b ) );
	s32		g( Interpolate( COL_GETG( cola ), COL_GETG( colb ), factor_g ) );
	s32		r( Interpolate( COL_GETR( cola ), COL_GETR( colb ), factor_r ) );

	return MAKE_COLOUR_U32( r, g, b, a );

}

inline u32	InterpolateColour( u32 cola, u32 colb, float factor )
{
	s32		a( Interpolate( COL_GETA( cola ), COL_GETA( colb ), factor ) );
	s32		b( Interpolate( COL_GETB( cola ), COL_GETB( colb ), factor ) );
	s32		g( Interpolate( COL_GETG( cola ), COL_GETG( colb ), factor ) );
	s32		r( Interpolate( COL_GETR( cola ), COL_GETR( colb ), factor ) );

	return MAKE_COLOUR_U32( r, g, b, a );
}




#endif
