/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __TEXTURECACHE_H__
#define __TEXTURECACHE_H__

#include "Texture.h"
#include "DaedCritSect.h"
#include "DaedSingleton.h"

struct TextureInfo
{
	u32	TmemAddress;
	u32	TmemPalAddress;

	u32 Address;		// Corresponds to dwAddress in Tile
	u32 Format;		// e.g. RGBA, IA
	u32 Size;		// e.g. 16bpp
	s32 Left;
	s32 Top;
	u32 Width;
	u32 Height;
	u32 Pitch;
	u32 TLutFmt;

	bool Swapped;			// Are odd lines word swapped?

	TextureInfo()
	{
		memset( this, 0, sizeof( TextureInfo ) );
	}

	u32 GetHashCode() const;
	//void DisplayDetails() const
	//{
	//	printf( "TmemAddress: %08x\n", TmemAddress );
	//	printf( "TmemPalAddress: %08x\n", TmemPalAddress );
	//	printf( "Address: %08x\n", Address );
	//	printf( "Format: %08x\n", Format );
	//	printf( "Size: %08x\n", Size );
	//	printf( "Left: %08x\n", Left );
	//	printf( "Top: %08x\n", Top );
	//	printf( "Width: %08x\n", Width );
	//	printf( "Height: %08x\n", Height );
	//	printf( "Pitch: %08x\n", Pitch );
	//	printf( "TLutFmt: %08x\n", TLutFmt );
	//	printf( "Swapped: %02x\n", Swapped );
	//}

	// or bool operator==(,,) ?
	BOOL IsEqual(const TextureInfo & ti) const
	{
		return ( memcmp(this, &ti, sizeof(TextureInfo)) == 0 );
	}

};

class CTextureCache : public CCritSect, public CSingleton< CTextureCache >
{
	public:
		virtual ~CTextureCache() {};

		virtual void PurgeOldTextures() = 0;
		virtual void DropTextures() = 0;

		virtual void SetDumpTextures( bool dump_textures ) = 0;
		virtual bool GetDumpTextures( ) const = 0;

		virtual void DisplayStats() = 0;

		virtual CTexture * GetTexture( TextureInfo * pti ) = 0;

		virtual bool InvalidateDeviceObjects() = 0;
};

#endif	// __TEXTURECACHE_H__
