/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "AboutScreen.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "ColourPulser.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\PSPColour.h"

#include "DaedMathUtil.h"

#include <pspctrl.h>
#include <pspgu.h>

namespace
{
	const u32				TEXT_AREA_TOP = 70;
	const u32				TEXT_AREA_LEFT = 40;
	const u32				TEXT_AREA_RIGHT = 480-40;
	const u32				TEXT_AREA_HEIGHT = 176;

	const u32				START_TEXT_TOP = 205;

	#ifdef _DEBUG
		const char * const DAEDALUS_VERSION_TEXT = "Daedalus PSP R6 (Debug Build)";
	#else
		const char * const DAEDALUS_VERSION_TEXT = "Daedalus PSP R6 (Release Build)";
	#endif

	const char * const		DATE_TEXT = "Built " __DATE__;

	const char * const INFO_TEXT = "For news and updates visit:";

	const char * const URL_TEXT_1 = "http://sf.net/projects/daedalus-n64/";
	const char * const URL_TEXT_2 = "http://strmnnrmn.blogspot.com/";


	const char * const PRESS_START_TEXT = "Press Start to Continue";

	const u32				MS_PER_COLOUR_CYCLE = 1200;

	const s32				MAX_TEXT_BRIGHTNESS( 0xee );

	const u32 COLOUR_SELECTION_DIM		= MAKE_COLOUR_U32( 0xe0, 0xe0, 0x80 );
	const u32 COLOUR_SELECTION_BRIGHT	= MAKE_COLOUR_U32( 0xff, 0xff, 0x80 );

	const u32 COLOUR_TEXT_BLUE			= MAKE_COLOUR_U32( 100,100, MAX_TEXT_BRIGHTNESS );
	const u32 COLOUR_TEXT_WHITE			= MAKE_COLOUR_U32( MAX_TEXT_BRIGHTNESS, MAX_TEXT_BRIGHTNESS, MAX_TEXT_BRIGHTNESS );

	const char * const		BACKGROUND_FILENAME = "Resources/background.png";
}

//*************************************************************************************
//
//*************************************************************************************
class IAboutScreen : public CAboutScreen, public CUIScreen
{
	public:

		IAboutScreen();
		~IAboutScreen();

		// CPspRomSelector
		virtual void				Run( CUIContext * p_context );

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mIsFinished; }

	private:
		bool						mIsFinished;
		CColourPulser				mColourPulser;
};

//*************************************************************************************
//
//*************************************************************************************
CAboutScreen::~CAboutScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
CAboutScreen *	CAboutScreen::Create()
{
	return new IAboutScreen;
}

//*************************************************************************************
//
//*************************************************************************************
IAboutScreen::IAboutScreen()
:	mIsFinished( false )
,	mColourPulser( COLOUR_SELECTION_DIM, COLOUR_SELECTION_BRIGHT, MS_PER_COLOUR_CYCLE )
{
}

//*************************************************************************************
//
//*************************************************************************************
IAboutScreen::~IAboutScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	IAboutScreen::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if((new_buttons & PSP_CTRL_START) ||
			(new_buttons & PSP_CTRL_CROSS))
		{
			mIsFinished = true;
		}
	}

	s32		elapsed_ms = s32( 1000.0f * elapsed_time );

	mColourPulser.Update( elapsed_ms );
}

//*************************************************************************************
//
//*************************************************************************************
void	IAboutScreen::Render()
{
	s32		x,y;
	const u32	font_height( CDrawText::GetFontHeight() );

	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, DAEDALUS_VERSION_TEXT, DrawTextUtilities::AT_CENTRE );
	y = TEXT_AREA_TOP;
	DrawTextUtilities::DrawShadowText( x, y, DAEDALUS_VERSION_TEXT, COLOUR_TEXT_WHITE );
	y += font_height;
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, DATE_TEXT, DrawTextUtilities::AT_CENTRE );
	DrawTextUtilities::DrawShadowText( x, y, DATE_TEXT, COLOUR_TEXT_WHITE );
	y += font_height;

	y += font_height;

	// URL 1
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, INFO_TEXT, DrawTextUtilities::AT_CENTRE );
	DrawTextUtilities::DrawShadowText( x, y, INFO_TEXT, COLOUR_TEXT_WHITE );
	y += font_height;


	y += font_height;

	// URL 1
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, URL_TEXT_1, DrawTextUtilities::AT_CENTRE );
	DrawTextUtilities::DrawShadowText( x, y, URL_TEXT_1, COLOUR_TEXT_BLUE );
	y += font_height;

	// URL 2
	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, URL_TEXT_2, DrawTextUtilities::AT_CENTRE );
	DrawTextUtilities::DrawShadowText( x, y, URL_TEXT_2, COLOUR_TEXT_BLUE );
	y += font_height;


	x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, PRESS_START_TEXT, DrawTextUtilities::AT_CENTRE );
	y = START_TEXT_TOP;

	DrawTextUtilities::DrawShadowText( x, y, PRESS_START_TEXT, mColourPulser.GetCurrentColour() );
}

//*************************************************************************************
//
//*************************************************************************************
void	IAboutScreen::Run( CUIContext * p_context )
{
	p_context->SetBackgroundTexture( BACKGROUND_FILENAME );
	p_context->Run( this );
}
