/*
Copyright (C) 2006 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "GlobalSettingsScreen.h"

#include "UIContext.h"
#include "UIScreen.h"
#include "ColourPulser.h"
#include "PSPGraphics\DrawText.h"
#include "PSPGraphics\PSPColour.h"

#include <pspctrl.h>
#include <pspgu.h>

namespace
{
	const u32		TITLE_AREA_TOP = 15;
	const u32		TITLE_AREA_LEFT = 16;
	const u32		TITLE_AREA_RIGHT = 480-16;

	const u32		TEXT_AREA_TOP = 15+16+16;
	const u32		TEXT_AREA_LEFT = 40;
	const u32		TEXT_AREA_RIGHT = 440;

	const u32		DESCRIPTION_AREA_TOP = 272-48-10;		// 3 rows
	const u32		DESCRIPTION_AREA_LEFT = 16;
	const u32		DESCRIPTION_AREA_RIGHT = 480-16;

	const char * const GLOBAL_SETTINGS_TEXT = "Settings";

	const u32		MS_PER_COLOUR_CYCLE = 1200;

	const u32		COLOUR_SELECTION_DIM	= MAKE_COLOUR_U32( 0xe0, 0xe0, 0x80 );
	const u32		COLOUR_SELECTION_BRIGHT	= MAKE_COLOUR_U32( 0xff, 0xff, 0x80 );

	const u32		COLOUR_TEXT_WHITE = 0xffeeeeee;
	const u32		COLOUR_TEXT_WHITE_DISABLED = 0xffd0d0d0;
	const u32		COLOUR_TEXT_RED = 0xff0000ff;
	const u32		COLOUR_TEXT_RED_DISABLED = 0xffb0b0ff;

	const char * const	BACKGROUND_FILENAME = "Resources/plain_background.png";

	enum	EGlobalSettingType
	{
		GS_TESSELATE_LARGE_TRIS = 0,
		GS_DISPLAY_FRAMERATE,

		NUM_GLOBAL_SETTINGS,
	};

	const EGlobalSettingType	RS_FIRST_SETTING = GS_TESSELATE_LARGE_TRIS;
	const EGlobalSettingType	RS_LAST_SETTING = GS_DISPLAY_FRAMERATE;

	struct SRomSettingInfo
	{
		const char *		Setting;
		const char *		Description;
	};

	const	SRomSettingInfo		gRomSettingInfo[] = 
	{
		{ "Tesselate Large Tris",	"Whether to break up large triangles when rendering. Disable this to improve framerate, at the expense of graphics quality." },
		{ "Display Framerate",		"Whether to show the framerate while the rom is running." },
	};

}

extern bool		gTesselateLargeTriangles;		// XXXX
extern bool		gDisplayFramerate;				// XXXX

//*************************************************************************************
//
//*************************************************************************************
class IGlobalSettingsScreen : public CGlobalSettingsScreen, public CUIScreen
{
	public:

		IGlobalSettingsScreen();
		~IGlobalSettingsScreen();

		// CPspRomSelettings
		virtual EScreenStatus		Run( CUIContext * p_context );

		// CUIScreen
		virtual void				Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons );
		virtual void				Render();
		virtual bool				IsFinished() const									{ return mIsFinished; }

				bool				IsSettingReadOnly( EGlobalSettingType global_setting ) const;
				void				NextSetting( EGlobalSettingType global_setting );
				void				PreviousSetting( EGlobalSettingType global_setting );
				const char *		GetSettingName( EGlobalSettingType global_setting ) const ;

	private:
		EScreenStatus				mScreenStatus;
		bool						mIsFinished;
		EGlobalSettingType			mSelectedSetting;

		CColourPulser				mColourPulser;
};

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsScreen::~CGlobalSettingsScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsScreen *	CGlobalSettingsScreen::Create()
{
	return new IGlobalSettingsScreen;
}

//*************************************************************************************
//
//*************************************************************************************
IGlobalSettingsScreen::IGlobalSettingsScreen()
:	mScreenStatus( SS_OK )
,	mIsFinished( false )
,	mSelectedSetting( RS_FIRST_SETTING )
,	mColourPulser( COLOUR_SELECTION_DIM, COLOUR_SELECTION_BRIGHT, MS_PER_COLOUR_CYCLE )
{
}

//*************************************************************************************
//
//*************************************************************************************
IGlobalSettingsScreen::~IGlobalSettingsScreen()
{
}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsScreen::Update( float elapsed_time, const v2 & stick, u32 old_buttons, u32 new_buttons )
{
	if(old_buttons != new_buttons)
	{
		if(new_buttons & PSP_CTRL_START)
		{
			mIsFinished = true;
			mScreenStatus = SS_OK;
		}
		if(new_buttons & PSP_CTRL_SELECT)
		{
			mIsFinished = true;
			mScreenStatus = SS_BACK;
		}

		if( new_buttons & PSP_CTRL_UP )
		{
			if( mSelectedSetting > RS_FIRST_SETTING )
			{
				mSelectedSetting = EGlobalSettingType( mSelectedSetting - 1 );
			}
		}
		if( new_buttons & PSP_CTRL_DOWN )
		{
			if( mSelectedSetting < RS_LAST_SETTING )
			{
				mSelectedSetting = EGlobalSettingType( mSelectedSetting + 1 );
			}
		}

		if( new_buttons & PSP_CTRL_LEFT )
		{
			if( !IsSettingReadOnly( mSelectedSetting ) )
			{
				PreviousSetting( mSelectedSetting );
			}
		}
		if( new_buttons & PSP_CTRL_RIGHT )
		{
			if( !IsSettingReadOnly( mSelectedSetting ) )
			{
				NextSetting( mSelectedSetting );
			}
		}
	}

	s32		elapsed_ms = s32( 1000.0f * elapsed_time );
	mColourPulser.Update( elapsed_ms );
}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsScreen::Render()
{
	s32		x,y;

	x = DrawTextUtilities::AlignText( TITLE_AREA_LEFT, TITLE_AREA_RIGHT, GLOBAL_SETTINGS_TEXT, DrawTextUtilities::AT_CENTRE );
	y = TITLE_AREA_TOP;

	DrawTextUtilities::DrawShadowText( x, y, GLOBAL_SETTINGS_TEXT, COLOUR_TEXT_RED );

	y = TEXT_AREA_TOP + 0;

	const u32	font_height( CDrawText::GetFontHeight() );

	for( u32 i = 0; i < NUM_GLOBAL_SETTINGS; ++i)
	{
		EGlobalSettingType	setting = EGlobalSettingType( i );

		bool				read_only( IsSettingReadOnly( setting ) );

		u32		colour;
		if( setting == mSelectedSetting )
		{
			colour = read_only ? COLOUR_TEXT_RED_DISABLED : mColourPulser.GetCurrentColour();
		}
		else
		{
			colour = read_only ? COLOUR_TEXT_WHITE_DISABLED : COLOUR_TEXT_WHITE;
		}

		x = TEXT_AREA_LEFT;
		DrawTextUtilities::DrawShadowText( x, y, gRomSettingInfo[i].Setting, colour );

		const char * p_setting( GetSettingName( setting ) );
		x = DrawTextUtilities::AlignText( TEXT_AREA_LEFT, TEXT_AREA_RIGHT, p_setting, DrawTextUtilities::AT_RIGHT );
		DrawTextUtilities::DrawShadowText( x, y, p_setting, colour );
		y += font_height;
	}

	x = DESCRIPTION_AREA_LEFT;
	y = DESCRIPTION_AREA_TOP;

	const char *		p_description( gRomSettingInfo[mSelectedSetting].Description );
	std::vector<u32>	lengths;
	DrawTextUtilities::WrapText( DESCRIPTION_AREA_RIGHT - DESCRIPTION_AREA_LEFT, p_description, strlen( p_description ), lengths );
	for( u32 i = 0; i < lengths.size(); ++i )
	{
		DrawTextUtilities::DrawShadowText( x, y, p_description, lengths[ i ], COLOUR_TEXT_WHITE );
		p_description += lengths[ i ];
		y += font_height;
	}
}

//*************************************************************************************
//
//*************************************************************************************
CGlobalSettingsScreen::EScreenStatus	IGlobalSettingsScreen::Run( CUIContext * p_context )
{
	p_context->SetBackgroundTexture( BACKGROUND_FILENAME );
	p_context->Run( this );

	return mScreenStatus;
}

//*************************************************************************************
//
//*************************************************************************************
bool	IGlobalSettingsScreen::IsSettingReadOnly( EGlobalSettingType global_setting ) const
{
	switch( global_setting )
	{
	case GS_TESSELATE_LARGE_TRIS:
	case GS_DISPLAY_FRAMERATE:
		return false;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		return false;
	}

}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsScreen::NextSetting( EGlobalSettingType global_setting )
{
	switch( global_setting )
	{
	case GS_TESSELATE_LARGE_TRIS:
		gTesselateLargeTriangles = !gTesselateLargeTriangles;
		break;

	case GS_DISPLAY_FRAMERATE:
		gDisplayFramerate = !gDisplayFramerate;
		break;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		break;
	}
}

//*************************************************************************************
//
//*************************************************************************************
void	IGlobalSettingsScreen::PreviousSetting( EGlobalSettingType global_setting )
{
	switch( global_setting )
	{
	case GS_TESSELATE_LARGE_TRIS:
		gTesselateLargeTriangles = !gTesselateLargeTriangles;
		break;

	case GS_DISPLAY_FRAMERATE:
		gDisplayFramerate = !gDisplayFramerate;
		break;

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		break;
	}
}

//*************************************************************************************
//
//*************************************************************************************
const char *	IGlobalSettingsScreen::GetSettingName( EGlobalSettingType global_setting ) const 
{
	switch( global_setting )
	{
	case GS_TESSELATE_LARGE_TRIS:
		return gTesselateLargeTriangles ? "Yes" : "No";

	case GS_DISPLAY_FRAMERATE:
		return gDisplayFramerate ? "Yes" : "No";

	default:
		DAEDALUS_ERROR( "Unhandled rom setting" );
		return "?";
	}
}

